"""Utility functions to help with testing Language Server features."""
import asyncio
import logging
import os
import pathlib
import threading
from concurrent.futures import Future
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import pygls.uris as Uri
from pygls.exceptions import JsonRpcMethodNotFound
from pygls.lsp.methods import CANCEL_REQUEST
from pygls.lsp.methods import COMPLETION
from pygls.lsp.methods import COMPLETION_ITEM_RESOLVE
from pygls.lsp.methods import DEFINITION
from pygls.lsp.methods import DOCUMENT_SYMBOL
from pygls.lsp.methods import EXIT
from pygls.lsp.methods import INITIALIZE
from pygls.lsp.methods import INITIALIZED
from pygls.lsp.methods import SHUTDOWN
from pygls.lsp.methods import TEXT_DOCUMENT_DID_CHANGE
from pygls.lsp.methods import TEXT_DOCUMENT_DID_CLOSE
from pygls.lsp.methods import TEXT_DOCUMENT_DID_OPEN
from pygls.lsp.methods import TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS
from pygls.lsp.methods import WINDOW_LOG_MESSAGE
from pygls.lsp.methods import WINDOW_SHOW_MESSAGE
from pygls.lsp.types import ClientCapabilities
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionList
from pygls.lsp.types import CompletionParams
from pygls.lsp.types import DefinitionParams
from pygls.lsp.types import Diagnostic
from pygls.lsp.types import DidChangeTextDocumentParams
from pygls.lsp.types import DidCloseTextDocumentParams
from pygls.lsp.types import DidOpenTextDocumentParams
from pygls.lsp.types import DocumentSymbol
from pygls.lsp.types import DocumentSymbolParams
from pygls.lsp.types import InitializeParams
from pygls.lsp.types import Location
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from pygls.lsp.types import TextDocumentContentChangeEvent
from pygls.lsp.types import TextDocumentIdentifier
from pygls.lsp.types import TextDocumentItem
from pygls.lsp.types import VersionedTextDocumentIdentifier
from pygls.protocol import LanguageServerProtocol
from pygls.server import LanguageServer
from sphinx import __version__ as __sphinx_version__

logger = logging.getLogger(__name__)


class ClientServer:
    """A class that returns a client server pair for use in tests.

    Originally based on:
    https://github.com/openlawlibrary/pygls/blob/704ff607a2a993a633fc2ad05eb95120dd34cad5/tests/conftest.py#L53
    """

    def __init__(self, server: LanguageServer):

        client_server_read, client_server_write = os.pipe()
        server_client_read, server_client_write = os.pipe()

        self.server = server
        self.server_thread = threading.Thread(
            name="Server Thread",
            target=self.server.start_io,
            args=(
                os.fdopen(client_server_read, "rb"),
                os.fdopen(server_client_write, "wb"),
            ),
        )
        self.server_thread.daemon = True

        self.client = new_test_client()
        self.client_thread = threading.Thread(
            name="Client Thread",
            target=self.client.start_io,
            args=(
                os.fdopen(server_client_read, "rb"),
                os.fdopen(client_server_write, "wb"),
            ),
        )
        self.client_thread.daemon = True

    async def start(
        self,
        root_uri: str,
        initialization_options: Optional[Dict[str, Any]] = None,
        wait=True,
    ):
        """Start the client and server components ready for testing.

        As well as starting the event loops for them both it also
        initializes the server based on the ``root_uri`` and
        ``initialization_options`` parameters. Once the response for
        the ``initialize`` request is received, it will proceed to send the
        ``initialized`` notification to trigger the initial build of the
        documentation.

        By default this method will ``await`` until the client receives the
        ``esbonio/buildComplete`` notification, though this can be disabled
        by passing ``wait=False`` (useful for preventing deadlocks in situations
        where the server is not expected to start successfully).

        Parameters
        ----------
        root_uri:
           The root uri of the project to initialize the server within
        initialization_options:
           The initializaion options to pass to the server
        wait:
           If ``True``, await until the client receives the ``esbonio/buildComplete``
           notification.
        """
        self.server_thread.start()
        self.client_thread.start()

        # Give the client and server some time to initialize
        while self.server.lsp.transport is None or self.client.lsp.transport is None:
            await asyncio.sleep(0.1)

        response = await self.client.lsp.send_request_async(
            INITIALIZE,
            InitializeParams(
                process_id=12345,
                root_uri=root_uri,
                capabilities=ClientCapabilities(),
                initialization_options=initialization_options or {},
            ),
        )

        assert "capabilities" in response

        self.client.lsp.notify(INITIALIZED)

        if wait:
            await self.client.lsp.wait_for_notification_async("esbonio/buildComplete")

        return

    async def stop(self):
        response = await self.client.lsp.send_request_async(SHUTDOWN)
        assert response is None

        self.client.lsp.notify(EXIT)
        self.server_thread.join()

        self.client._stop_event.set()
        try:
            self.client.loop._signal_handlers.clear()
        except AttributeError:
            pass

        self.client_thread.join()


def sphinx_version(eq: Optional[int] = None) -> bool:
    """Helper function for determining which version of Sphinx we are
    testing with.

    Currently this only cares about the major version number.

    Parameters
    ----------
    eq:
       When set returns ``True`` if the Sphinx version exactly matches
       what's given.
    """

    major, _, _ = [int(v) for v in __sphinx_version__.split(".")]

    if eq and major == eq:
        return True

    return False


def directive_argument_patterns(name: str, partial: str = "") -> List[str]:
    """Return a number of example directive argument patterns.

    These correspond to test cases where directive argument suggestions should be
    generated.

    Parameters
    ----------
    name:
       The name of the directive to generate suggestions for.
    partial:
       The partial argument that the user has already entered.
    """
    return [s.format(name, partial) for s in [".. {}:: {}", "   .. {}:: {}"]]


def role_patterns(partial: str = "") -> List[str]:
    """Return a number of example role patterns.

    These correspond to when role suggestions should be generated.

    Parameters
    ----------
    partial:
       The partial role name that the user has already entered
    """
    return [
        s.format(partial)
        for s in [
            "{}",
            "({}",
            "- {}",
            "   {}",
            "   ({}",
            "   - {}",
            "some text {}",
            "some text ({}",
            "   some text {}",
            "   some text ({}",
        ]
    ]


def role_target_patterns(
    name: str, partial: str = "", include_modifiers: bool = True
) -> List[str]:
    """Return a number of example role target patterns.

    These correspond to test cases where role target suggestions should be generated.

    Parameters
    ----------
    name:
       The name of the role to generate suggestions for.
    partial:
       The partial target that the user as already entered.
    include_modifiers:
       A flag to indicate if additional modifiers like ``!`` and ``~`` should be
       included in the generated patterns.
    """

    patterns = [
        ":{}:`{}",
        "(:{}:`{}",
        "- :{}:`{}",
        ":{}:`More Info <{}",
        "(:{}:`More Info <{}",
        "   :{}:`{}",
        "   (:{}:`{}",
        "   - :{}:`{}",
        "   :{}:`Some Label <{}",
        "   (:{}:`Some Label <{}",
    ]

    test_cases = [p.format(name, partial) for p in patterns]

    if include_modifiers:
        test_cases += [p.format(name, "!" + partial) for p in patterns]
        test_cases += [p.format(name, "~" + partial) for p in patterns]

    return test_cases


def intersphinx_target_patterns(name: str, project: str) -> List[str]:
    """Return a number of example intersphinx target patterns.

    These correspond to cases where target completions may be generated

    Parameters
    ----------
    name: str
       The name of the role to generate examples for
    project: str
       The name of the project to generate examples for
    """
    return [
        s.format(name, project)
        for s in [
            ":{}:`{}:",
            "(:{}:`{}:",
            ":{}:`More Info <{}:",
            "(:{}:`More Info <{}:",
            "   :{}:`{}:",
            "   (:{}:`{}:",
            "   :{}:`Some Label <{}:",
            "   (:{}:`Some Label <{}:",
        ]
    ]


async def completion_request(
    test: ClientServer, test_uri: str, text: str, character: Optional[int] = None
) -> CompletionList:
    """Make a completion request to a language server.

    Intended for use within test cases, this function simulates the opening of a
    document, inserting some text, triggering a completion request and closing it
    again.

    The file referenced by ``test_uri`` does not have to exist.

    The text to be inserted is specified through the ``text`` parameter. By default
    it's assumed that the ``text`` parameter consists of a single line of text, in fact
    this function will error if that is not the case.

    If your request requires additional context (such as directive option completions)
    it can be included but it must be delimited with a ``\\f`` character. For example,
    to represent the following scenario::

       .. image:: filename.png
          :align: center
          :
           ^

    where ``^`` represents the position from which we trigger the completion request.
    We would set ``text`` to the following
    ``.. image:: filename.png\\n   :align: center\\n\\f   :``

    Parameters
    ----------
    test:
       The client-server pair to be used to make the request.
    test_uri:
       The uri the completion request should be made within.
    text
       The text that provides the context for the completion request.
    character:
       The character index at which to make the completion request from.
       If ``None``, it will default to the end of the inserted text.
    """

    if "\f" in text:
        contents, text = text.split("\f")
    else:
        contents = ""

    logger.debug("Context text:    '%s'", contents)
    logger.debug("Insertion text: '%s'", text)
    assert "\n" not in text, "Insertion text cannot contain newlines"

    ext = pathlib.Path(Uri.to_fs_path(test_uri)).suffix
    lang_id = "python" if ext == ".py" else "rst"

    test.client.lsp.notify(
        TEXT_DOCUMENT_DID_OPEN,
        DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=test_uri, language_id=lang_id, version=1, text=contents
            )
        ),
    )

    lines = contents.split("\n")
    line = len(lines) - 1
    insertion_point = len(lines[-1])

    logger.debug("Insertion pos: (%d, %d)", line, insertion_point)

    test.client.lsp.notify(
        TEXT_DOCUMENT_DID_CHANGE,
        DidChangeTextDocumentParams(
            text_document=VersionedTextDocumentIdentifier(uri=test_uri, version=2),
            content_changes=[
                TextDocumentContentChangeEvent(
                    range=Range(
                        start=Position(line=line, character=insertion_point),
                        end=Position(line=line, character=insertion_point + len(text)),
                    ),
                    text=text,
                )
            ],
        ),
    )

    character = character or insertion_point + len(text)
    response = await test.client.lsp.send_request_async(
        COMPLETION,
        CompletionParams(
            text_document=TextDocumentIdentifier(uri=test_uri),
            position=Position(line=line, character=character),
        ),
    )

    test.client.lsp.notify(
        TEXT_DOCUMENT_DID_CLOSE,
        DidCloseTextDocumentParams(text_document=TextDocumentIdentifier(uri=test_uri)),
    )

    return CompletionList(**response)


async def completion_resolve_request(
    test: ClientServer, item: CompletionItem
) -> CompletionItem:
    """Make a ``completionItem/resolve`` request to a language server.

    Intented for use within test cases, this function assumes that a
    ``textDocument/completion`` request has already been performed resulting in a list
    of ``CompletionItem`` candidates being generated. This function can then be used
    to simulate the user selecting an item and the client making a follow up
    ``completionItem/resolve`` request to the server.
    """
    response = await test.client.lsp.send_request_async(COMPLETION_ITEM_RESOLVE, item)
    return CompletionItem(**response)


async def definition_request(
    test: ClientServer, test_uri: str, position: Position
) -> List[Location]:
    """Make a definition request to a language server.

    Intended for use within test cases, this function simulates opening an
    existing file and making a definition request and closing it again.

    The file referenced by ``test_uri`` *must* exist as the actual content of the file
    will be used by the language server.

    Parameters
    ----------
    test:
       The client-server pair to be used to make the request
    test_uri:
       The uri of the file the definition request should be made within.
    position:
       The position at which to trigger the completion request.
    """

    path = pathlib.Path(Uri.to_fs_path(test_uri))
    with path.open() as f:
        contents = f.read()

    test.client.lsp.notify(
        TEXT_DOCUMENT_DID_OPEN,
        DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=test_uri, language_id="rst", version=1, text=contents
            )
        ),
    )

    response = await test.client.lsp.send_request_async(
        DEFINITION,
        DefinitionParams(
            text_document=TextDocumentIdentifier(uri=test_uri), position=position
        ),
    )

    return [Location(**obj) for obj in response]


async def document_symbols_request(
    test: ClientServer, test_uri: str
) -> List[DocumentSymbol]:
    """Make a ``textDocument/documentSymbols`` request to a language server.

    Intended for use with test cases, this function simulates opening an existing
    file, making a document symbols request and closing it again.

    The file referenced by ``test_uri`` *must* exist as the content of the file will
    be used by the language server.

    Parameters
    ----------
    test:
       The client-server pair to be used
    test_uri:
       The uri of the file to make the ``textDocument/documentSymbols`` request within
    """

    path = pathlib.Path(Uri.to_fs_path(test_uri))
    with path.open() as f:
        contents = f.read()

    test.client.lsp.notify(
        TEXT_DOCUMENT_DID_OPEN,
        DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=test_uri, language_id="rst", version=1, text=contents
            )
        ),
    )

    response = await test.client.lsp.send_request_async(
        DOCUMENT_SYMBOL,
        DocumentSymbolParams(text_document=TextDocumentIdentifier(uri=test_uri)),
    )

    return [DocumentSymbol(**obj) for obj in response]


class ClientProtocol(LanguageServerProtocol):
    """A slightly modified version of the base protocol that allows
    us to await on notifications."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._notification_futures = {}

    def _handle_notification(self, method_name, params):

        if method_name == CANCEL_REQUEST:
            self._handle_cancel_notification(params.id)
            return

        future = self._notification_futures.pop(method_name, None)
        if future:
            future.set_result(params)

        try:
            handler = self._get_handler(method_name)
            self._execute_notification(handler, params)
        except (KeyError, JsonRpcMethodNotFound):
            logger.warning('Ignoring notification for unknown method "%s"', method_name)
        except Exception:
            logger.exception(
                'Failed to handle notification "%s": %s', method_name, params
            )

    def wait_for_notification(self, method, callback=None):

        future = Future()
        if callback:

            def wrapper(future: Future):
                result = future.result()
                callback(result)

            future.add_done_callback(wrapper)

        self._notification_futures[method] = future
        return future

    def wait_for_notification_async(self, method):
        fut = self.wait_for_notification(method)
        return asyncio.wrap_future(fut)


class Client(LanguageServer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.messages: List[Any] = []
        self.diagnostics: Dict[str, List[Diagnostic]] = {}


def new_test_client() -> Client:

    client = Client(protocol_cls=ClientProtocol, loop=asyncio.new_event_loop())

    @client.feature(TEXT_DOCUMENT_PUBLISH_DIAGNOSTICS)
    def publish_diagnostics(client: Client, params):
        client.diagnostics[params.uri] = params.diagnostics

    @client.feature(WINDOW_LOG_MESSAGE)
    def log_message(client: Client, params):
        pass

    @client.feature(WINDOW_SHOW_MESSAGE)
    def show_message(client: Client, params):
        client.messages.append(params)

    @client.feature("esbonio/buildStart")
    def build_start(client: Client, params):
        pass

    @client.feature("esbonio/buildComplete")
    def build_complete(client: Client, params):
        pass

    return client



"""Support for Sphinx domains."""
import pathlib
import typing
from typing import List
from typing import Optional
from typing import Set

import pygls.uris as Uri
from docutils import nodes
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionItemKind
from pygls.lsp.types import Location
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from pygls.workspace import Document
from sphinx.domains import Domain

from esbonio.lsp.roles import Roles
from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.rst import DefinitionContext
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.sphinx import SphinxLanguageServer


TARGET_KINDS = {
    "attribute": CompletionItemKind.Field,
    "doc": CompletionItemKind.File,
    "class": CompletionItemKind.Class,
    "envvar": CompletionItemKind.Variable,
    "function": CompletionItemKind.Function,
    "method": CompletionItemKind.Method,
    "module": CompletionItemKind.Module,
    "term": CompletionItemKind.Text,
}


class DomainFeatures:
    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)

    def complete_targets(
        self, context: CompletionContext, name: str, domain: Optional[str]
    ) -> List[CompletionItem]:

        groups = context.match.groupdict()
        domain = domain or ""
        label = groups["label"]

        if ":" in label:
            return self.complete_intersphinx_targets(name, domain, label)

        items = [
            object_to_completion_item(o)
            for o in self.rst.get_role_targets(name, domain)
        ]

        for project in self.rst.get_intersphinx_projects():
            if self.rst.has_intersphinx_targets(project, name, domain):
                items.append(project_to_completion_item(project))

        return items

    def complete_intersphinx_targets(
        self, name: str, domain: str, label: str
    ) -> List[CompletionItem]:
        items = []
        project, *_ = label.split(":")
        intersphinx_targets = self.rst.get_intersphinx_targets(project, name, domain)

        for type_, targets in intersphinx_targets.items():
            items += [
                intersphinx_target_to_completion_item(project, label, target, type_)
                for label, target in targets.items()
            ]

        return items

    def find_definitions(
        self, context: DefinitionContext, name: str, domain: Optional[str]
    ) -> List[Location]:

        label = context.match.group("label")

        if not domain and name == "ref":
            return self.ref_definition(label)

        if not domain and name == "doc":
            return self.doc_definition(context.doc, label)

        return []

    def doc_definition(self, doc: Document, label: str) -> List[Location]:
        """Goto definition implementation for ``:doc:`` targets"""

        if self.rst.app is None:
            return []

        srcdir = self.rst.app.srcdir
        currentdir = pathlib.Path(Uri.to_fs_path(doc.uri)).parent

        if label.startswith("/"):
            path = str(pathlib.Path(srcdir, label[1:] + ".rst"))
        else:
            path = str(pathlib.Path(currentdir, label + ".rst"))

        return [
            Location(
                uri=Uri.from_fs_path(path),
                range=Range(
                    start=Position(line=0, character=0),
                    end=Position(line=1, character=0),
                ),
            )
        ]

    def ref_definition(self, label: str) -> List[Location]:
        """Goto definition implementation for ``:ref:`` targets"""

        types = set(self.rst.get_role_target_types("ref"))
        std = self.rst.get_domain("std")
        if std is None:
            return []

        docname = self.find_docname_for_label(label, std, types)
        if docname is None:
            return []

        doctree = self.rst.get_doctree(docname=docname)
        if doctree is None:
            return []

        uri = None
        line = None

        for node in doctree.traverse(condition=nodes.target):

            if "refid" not in node:
                continue

            if label == node["refid"].replace("-", "_"):
                uri = Uri.from_fs_path(node.source)
                line = node.line
                break

        if uri is None or line is None:
            return []

        return [
            Location(
                uri=uri,
                range=Range(
                    start=Position(line=line - 1, character=0),
                    end=Position(line=line, character=0),
                ),
            )
        ]

    def find_docname_for_label(
        self, label: str, domain: Domain, types: Optional[Set[str]] = None
    ) -> Optional[str]:
        """Given the label name and domain it belongs to, return the docname its
        definition resides in.

        Parameters
        ----------
        label:
           The label to search for
        domain:
           The domain to search within
        types:
           A collection of object types that the label chould have.
        """

        docname = None
        types = types or set()

        # _, title, _, _, anchor, priority
        for name, _, type_, doc, _, _ in domain.get_objects():
            if types and type_ not in types:
                continue

            if name == label:
                docname = doc
                break

        return docname


def intersphinx_target_to_completion_item(
    project: str, label: str, target: tuple, type_: str
) -> CompletionItem:

    # _. _. url, _
    source, version, _, display = target

    display_name = label if display == "-" else display
    completion_kind = ":".join(type_.split(":")[1:]) if ":" in type_ else type_

    if version:
        version = f" v{version}"

    return CompletionItem(
        label=label,
        detail=f"{display_name} - {source}{version}",
        kind=TARGET_KINDS.get(completion_kind, CompletionItemKind.Reference),
        insert_text=f"{project}:{label}",
    )


def object_to_completion_item(object_: tuple) -> CompletionItem:

    # _, _, _, docname, anchor, priority
    name, display_name, type_, _, _, _ = object_
    insert_text = name

    key = type_.split(":")[1] if ":" in type_ else type_
    kind = TARGET_KINDS.get(key, CompletionItemKind.Reference)

    # ensure :doc: targets are inserted as an absolute path - that way the reference
    # will always work regardless of the file's location.
    if type_ == "doc":
        insert_text = f"/{name}"

    # :option: targets need to be inserted as `<progname> <option>` in order to resolve
    # correctly. However, this only seems to be the case "locally" as
    # `<progname>.<option>` seems to resolve fine when using intersphinx...
    if type_ == "cmdoption":
        name = " ".join(name.split("."))
        display_name = name
        insert_text = name

    return CompletionItem(
        label=name, kind=kind, detail=str(display_name), insert_text=insert_text
    )


def project_to_completion_item(project: str) -> CompletionItem:
    return CompletionItem(
        label=project, detail="intersphinx", kind=CompletionItemKind.Module
    )


def esbonio_setup(rst: RstLanguageServer):

    if isinstance(rst, SphinxLanguageServer):
        domains = DomainFeatures(rst)
        roles = rst.get_feature("esbonio.lsp.roles.Roles")

        if roles:
            typing.cast(Roles, roles).add_target_definition_provider(domains)
            typing.cast(Roles, roles).add_target_completion_provider(domains)



import collections
import importlib
import logging
import pathlib
import re
import typing
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import docutils.parsers.rst.directives as directives  # type: ignore
import pygls.uris as Uri
from docutils import nodes
from docutils.nodes import NodeVisitor
from docutils.parsers.rst import Directive
from docutils.parsers.rst import roles
from pydantic import BaseModel
from pydantic import Field
from pygls import IS_WIN
from pygls.lsp.types import CodeAction
from pygls.lsp.types import CodeActionParams
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import DeleteFilesParams
from pygls.lsp.types import Diagnostic
from pygls.lsp.types import DidSaveTextDocumentParams
from pygls.lsp.types import DocumentSymbol
from pygls.lsp.types import InitializedParams
from pygls.lsp.types import InitializeParams
from pygls.lsp.types import Location
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from pygls.lsp.types import SymbolKind
from pygls.server import LanguageServer
from pygls.workspace import Document

from esbonio.cli import setup_cli


TRIPLE_QUOTE = re.compile("(\"\"\"|''')")
"""A regular expression matching the triple quotes used to delimit python docstrings."""

DEFAULT_MODULES = [
    "esbonio.lsp.directives",
    "esbonio.lsp.roles",
]
"""The modules to load in the default configuration of the server."""


class CompletionContext:
    """A class that captures the context within which a completion request has been
    made."""

    def __init__(
        self, *, doc: Document, location: str, match: "re.Match", position: Position
    ):

        self.doc = doc
        """The document within which the completion request was made."""

        self.location = location
        """The location type where the request was made.
        See :meth:`~esbonio.lsp.rst.RstLanguageServer.get_location_type` for details."""

        self.match = match
        """The match object describing the site of the completion request."""

        self.position = position
        """The position at which the completion request was made."""

    def __repr__(self):
        p = f"{self.position.line}:{self.position.character}"
        return (
            f"CompletionContext<{self.doc.uri}:{p} ({self.location}) -- {self.match}>"
        )


class DefinitionContext:
    """A class that captures the context within which a definition request has been
    made."""

    def __init__(
        self, *, doc: Document, location: str, match: "re.Match", position: Position
    ):

        self.doc = doc
        """The document within which the definition request was made."""

        self.location = location
        """The location type where the request was made.
        See :meth:`~esbonio.lsp.rst.RstLanguageServer.get_location_type` for details."""

        self.match = match
        """The match object describing the site of the definition request."""

        self.position = position
        """The position at which the definition request was made."""

    def __repr__(self):
        p = f"{self.position.line}:{self.position.character}"
        return (
            f"DefinitionContext<{self.doc.uri}:{p} ({self.location}) -- {self.match}>"
        )


class LanguageFeature:
    """Base class for language features."""

    def __init__(self, rst: "RstLanguageServer"):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)

    def initialize(self, options: InitializeParams) -> None:
        """Called once when the server is first initialized."""

    def initialized(self, params: InitializedParams) -> None:
        """Called once upon receipt of the `initialized` notification from the client."""

    def save(self, params: DidSaveTextDocumentParams) -> None:
        """Called each time a document is saved."""

    def delete_files(self, params: DeleteFilesParams) -> None:
        """Called each time files are deleted."""

    def code_action(self, params: CodeActionParams) -> List[CodeAction]:
        """Called when code actions should be computed."""
        return []

    completion_triggers: List["re.Pattern"] = []
    """A list of regular expressions used to determine if the
    :meth`~esbonio.lsp.rst.LanguageFeature.complete` method should be called on the
    current line."""

    def complete(self, context: CompletionContext) -> List[CompletionItem]:
        """Called if any of the given ``completion_triggers`` match the current line.

        This method should return a list of ``CompletionItem`` objects.

        Parameters
        ----------
        context:
           The context of the completion request
        """
        return []

    def completion_resolve(self, item: CompletionItem) -> CompletionItem:
        """Called with a completion item the user has selected in the UI,
        allowing for additional details to be provided e.g. documentation.

        Parameters
        ----------
        item:
            The completion item to provide additional details for.
        """
        return item

    definition_triggers: List["re.Pattern"] = []
    """A list of regular expressions used to determine if the
    :meth:`~esbonio.lsp.rst.LanguageFeature.definition` method should be called."""

    def definition(self, context: DefinitionContext) -> List[Location]:
        """Called if any of the given ``definition_triggers`` match the current line.

        This method should return a list of ``Location`` objects.

        Parameters
        ----------
        context:
           The context of the definition request.
        """
        return []


class DiagnosticList(collections.UserList):
    """A list type dedicated to holding diagnostics.

    This is mainly to ensure that only one instance of a diagnostic ever gets
    reported.
    """

    def append(self, item: Diagnostic):

        if not isinstance(item, Diagnostic):
            raise TypeError("Expected Diagnostic")

        for existing in self.data:
            fields = [
                existing.range == item.range,
                existing.message == item.message,
                existing.severity == item.severity,
                existing.code == item.code,
                existing.source == item.source,
            ]

            if all(fields):
                # Item already added, nothing to do.
                return

        self.data.append(item)


class ServerConfig(BaseModel):
    """Configuration options for the server."""

    log_level: str = Field("error", alias="logLevel")
    """The logging level of server messages to display."""

    log_filter: List[str] = Field(default_factory=list, alias="logFilter")
    """A list of logger names to restrict output to."""


class InitializationOptions(BaseModel):
    """The initialization options we can expect to receive from a client."""

    server: ServerConfig = Field(default_factory=ServerConfig)
    """The ``esbonio.server.*`` namespace of options."""


class RstLanguageServer(LanguageServer):
    """A generic reStructuredText language server."""

    def __init__(self, logger: Optional[logging.Logger] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.logger = logger or logging.getLogger(__name__)
        """The base logger that should be used for all language sever log entries."""

        self.user_config: Optional[BaseModel] = None
        """The user's configuration."""

        self._diagnostics: Dict[Tuple[str, str], List[Diagnostic]] = {}
        """Where we store and manage diagnostics."""

        self._features: Dict[str, LanguageFeature] = {}
        """The list of language features registered with the server."""

        self._directives: Optional[Dict[str, Directive]] = None
        """Cache for known directives."""

        self._roles: Optional[Dict[str, Any]] = None
        """Cache for known roles."""

    @property
    def configuration(self) -> Dict[str, Any]:
        """Return the server's actual configuration."""
        if not self.user_config:
            return {}

        return self.user_config.dict()

    def initialize(self, params: InitializeParams):
        self.user_config = InitializationOptions(
            **typing.cast(Dict, params.initialization_options)
        )
        self._configure_logging(self.user_config.server)

    def initialized(self, params: InitializedParams):
        pass

    def save(self, params: DidSaveTextDocumentParams):
        pass

    def delete_files(self, params: DeleteFilesParams):
        pass

    def add_feature(self, feature: "LanguageFeature"):
        """Register a language feature with the server."""

        key = f"{feature.__module__}.{feature.__class__.__name__}"
        self._features[key] = feature

    def get_feature(self, key) -> Optional["LanguageFeature"]:
        return self._features.get(key, None)

    def get_doctree(
        self, *, docname: Optional[str] = None, uri: Optional[str] = None
    ) -> Optional[Any]:
        return None

    def get_directives(self) -> Dict[str, Directive]:
        """Return a dictionary of the known directives"""

        if self._directives is not None:
            return self._directives

        ignored_directives = ["restructuredtext-test-directive"]
        found_directives = {**directives._directive_registry, **directives._directives}

        self._directives = {
            k: resolve_directive(v)
            for k, v in found_directives.items()
            if k not in ignored_directives
        }

        return self._directives

    def get_directive_options(self, name: str) -> Dict[str, Any]:
        """Return the options specification for the given directive."""

        directive = self.get_directives().get(name, None)
        if directive is None:
            return {}

        return directive.option_spec or {}

    def get_roles(self) -> Dict[str, Any]:
        """Return a dictionary of known roles."""

        if self._roles is not None:
            return self._roles

        found_roles = {**roles._roles, **roles._role_registry}

        self._roles = {
            k: v for k, v in found_roles.items() if v != roles.unimplemented_role
        }

        return self._roles

    def get_default_role(self) -> Tuple[Optional[str], Optional[str]]:
        """Return the default role for the project."""
        return None, None

    def clear_diagnostics(self, source: str, uri: Optional[str] = None) -> None:
        """Clear diagnostics from the given source.

        Parameters
        ----------
        source:
           The source from which to clear diagnostics.
        uri:
           If given, clear diagnostics from within just this uri. Otherwise, all
           diagnostics from the given source are cleared.
        """

        if uri:
            uri = normalise_uri(uri)

        for key in self._diagnostics.keys():
            clear_source = source == key[0]
            clear_uri = uri == key[1] or uri is None

            if clear_source and clear_uri:
                self._diagnostics[key] = []

    def set_diagnostics(
        self, source: str, uri: str, diagnostics: List[Diagnostic]
    ) -> None:
        """Set the diagnostics for the given source and uri.

        Parameters
        ----------
        source:
           The source the diagnostics are from
        uri:
           The uri the diagnostics are associated with
        diagnostics:
           The diagnostics themselves
        """
        uri = normalise_uri(uri)
        self._diagnostics[(source, uri)] = diagnostics

    def sync_diagnostics(self) -> None:
        """Update the client with the currently stored diagnostics."""

        uris = {uri for _, uri in self._diagnostics.keys()}
        diagnostics = {uri: DiagnosticList() for uri in uris}

        for (source, uri), diags in self._diagnostics.items():
            for diag in diags:
                diag.source = source
                diagnostics[uri].append(diag)

        for uri, diag_list in diagnostics.items():
            self.logger.debug("Publishing %d diagnostics for: %s", len(diag_list), uri)
            self.publish_diagnostics(uri, diag_list.data)

    def get_location_type(self, doc: Document, position: Position) -> str:
        """Given a document and a position, return the kind of location that
        represents.

        This will return one of the following values:

        - ``rst``: Indicates that the position is within an ``*.rst`` document
        - ``py``: Indicates that the position is within code in a ``*.py`` document
        - ``docstring``: Indicates that the position is within a docstring in a
          ``*.py`` document.

        Parameters
        ----------
        doc:
           The document associated with the given position
        position:
           The position to determine the type of.
        """
        ext = pathlib.Path(Uri.to_fs_path(doc.uri)).suffix

        if ext == ".rst":
            return "rst"

        if ext == ".py":

            # Let's count how many pairs of triple quotes are "above" us in the file
            # even => we're outside a docstring
            # odd  => we're within a docstring
            source = self.text_to_position(doc, position)
            count = len(TRIPLE_QUOTE.findall(source))
            return "py" if count % 2 == 0 else "docstring"

        # Fallback to rst
        self.logger.debug("Unable to determine location type for uri: %s", doc.uri)
        return "rst"

    def line_at_position(self, doc: Document, position: Position) -> str:
        """Return the contents of the line corresponding to the given position.

        Parameters
        ----------
        doc:
           The document associated with the given position
        position:
           The position representing the line to retrieve
        """

        try:
            return doc.lines[position.line]
        except IndexError:
            return ""

    def line_to_position(self, doc: Document, position: Position) -> str:
        """Return the contents of the line up until the given position.

        Parameters
        ----------
        doc:
           The document associated with the given position.
        position:
           The position representing the line to retrieve.
        """

        line = self.line_at_position(doc, position)
        return line[: position.character]

    def text_to_position(self, doc: Document, position: Position) -> str:
        """Return the contents of the document up until the given position.

        Parameters
        ----------
        doc:
           The document associated with the given position
        position:
           The position representing the point at which to stop gathering text.
        """
        idx = doc.offset_at_position(position)
        return doc.source[:idx]

    def _configure_logging(self, config: ServerConfig):

        level = LOG_LEVELS[config.log_level]

        lsp_logger = logging.getLogger("esbonio.lsp")
        lsp_logger.setLevel(level)

        lsp_handler = LspHandler(self)
        lsp_handler.setLevel(level)

        if len(config.log_filter) > 0:
            lsp_handler.addFilter(LogFilter(config.log_filter))

        formatter = logging.Formatter("[%(name)s] %(message)s")
        lsp_handler.setFormatter(formatter)
        lsp_logger.addHandler(lsp_handler)


class SymbolVisitor(NodeVisitor):
    """A visitor used to build the hierarchy we return from a
    ``textDocument/documentSymbol`` request.

    Currently this only looks at sections.
    """

    def __init__(self, rst, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.logger = rst.logger
        self.symbols = []
        self.symbol_stack = []

    @property
    def current_symbol(self) -> Optional[DocumentSymbol]:

        if len(self.symbol_stack) == 0:
            return None

        return self.symbol_stack[-1]

    def push_symbol(self):
        symbol = DocumentSymbol(
            name="",
            kind=SymbolKind.String,
            range=Range(
                start=Position(line=1, character=0),
                end=Position(line=1, character=10),
            ),
            selection_range=Range(
                start=Position(line=1, character=0),
                end=Position(line=1, character=10),
            ),
            children=[],
        )
        current_symbol = self.current_symbol

        if not current_symbol:
            self.symbols.append(symbol)
        else:
            current_symbol.children.append(symbol)

        self.symbol_stack.append(symbol)

    def pop_symbol(self):
        self.symbol_stack.pop()

    def visit_section(self, node: nodes.Node) -> None:
        self.push_symbol()

    def depart_section(self, node: nodes.Node) -> None:
        self.pop_symbol()

    def visit_title(self, node: nodes.Node) -> None:

        symbol = self.current_symbol
        if not symbol:
            raise ValueError("Missing expected current symbol")

        name = node.astext()
        line = (node.line or 1) - 1

        symbol.name = name
        symbol.range.start.line = line
        symbol.range.end.line = line
        symbol.range.end.character = len(name) - 1
        symbol.selection_range.start.line = line
        symbol.selection_range.end.line = line
        symbol.selection_range.end.character = len(name) - 1

    def depart_title(self, node: nodes.Node) -> None:
        pass

    def visit_Text(self, node: nodes.Node) -> None:
        pass

    def depart_Text(self, node: nodes.Node) -> None:
        pass

    def unknown_visit(self, node: nodes.Node) -> None:
        pass

    def unknown_departure(self, node: nodes.Node) -> None:
        pass


LOG_LEVELS = {
    "debug": logging.DEBUG,
    "error": logging.ERROR,
    "info": logging.INFO,
}


class LogFilter(logging.Filter):
    """A log filter that accepts message from any of the listed logger names."""

    def __init__(self, names):
        self.names = names

    def filter(self, record):
        return any(record.name == name for name in self.names)


class LspHandler(logging.Handler):
    """A logging handler that will send log records to an LSP client."""

    def __init__(self, server: RstLanguageServer):
        super().__init__()
        self.server = server

    def emit(self, record: logging.LogRecord) -> None:
        """Sends the record to the client."""

        # To avoid infinite recursions, it's simpler to just ignore all log records
        # coming from pygls...
        if "pygls" in record.name:
            return

        log = self.format(record).strip()
        self.server.show_message_log(log)


def resolve_directive(directive: Union[Directive, Tuple[str, str]]) -> Directive:
    """Return the directive based on the given reference.

    'Core' docutils directives are returned as tuples ``(modulename, ClassName)``
    so they need to be resolved manually.
    """

    if isinstance(directive, tuple):
        mod, cls = directive

        modulename = "docutils.parsers.rst.directives.{}".format(mod)
        module = importlib.import_module(modulename)
        return getattr(module, cls)

    return directive


def normalise_uri(uri: str) -> str:

    uri = Uri.from_fs_path(Uri.to_fs_path(uri))

    # Paths on windows are case insensitive.
    if IS_WIN:
        uri = uri.lower()

    return uri


cli = setup_cli("esbonio.lsp.rst", "Esbonio's reStructuredText language server.")
cli.set_defaults(modules=DEFAULT_MODULES)
cli.set_defaults(server_cls=RstLanguageServer)



"""Support for code-blocks and related directives."""
import textwrap
import typing
from typing import List

from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionItemKind
from pygls.lsp.types import MarkupContent
from pygls.lsp.types import MarkupKind
from pygments.lexers import get_all_lexers

from esbonio.lsp.directives import Directives
from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.sphinx import SphinxLanguageServer


class CodeBlocks:
    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)
        self._lexers: List[CompletionItem] = []

    @property
    def lexers(self) -> List[CompletionItem]:

        if len(self._lexers) == 0:
            self._index_lexers()

        return self._lexers

    def complete_arguments(
        self, context: CompletionContext, domain: str, name: str
    ) -> List[CompletionItem]:

        if name in {"code-block", "highlight"}:
            return self.lexers

        return []

    def _index_lexers(self):
        self._lexers = []

        for (name, labels, files, mimes) in get_all_lexers():
            for label in labels:
                documentation = f"""\
                ### {name}

                Filenames:  {', '.join(files)}

                MIME Types: {', '.join(mimes)}
                """

                item = CompletionItem(
                    label=label,
                    kind=CompletionItemKind.Constant,
                    documentation=MarkupContent(
                        kind=MarkupKind.Markdown, value=textwrap.dedent(documentation)
                    ),
                )

                self._lexers.append(item)


def esbonio_setup(rst: RstLanguageServer):

    name = "esbonio.lsp.directives.Directives"
    directives = rst.get_feature(name)

    if not isinstance(rst, SphinxLanguageServer) or not directives:
        return

    # To keep mypy happy.
    directives = typing.cast(Directives, directives)
    directives.add_argument_completion_provider(CodeBlocks(rst))



import hashlib
import logging
import pathlib
import re
import traceback
import typing
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import appdirs
import pygls.uris as Uri
from docutils.parsers.rst import Directive
from pydantic import BaseModel
from pydantic import Field
from pygls import IS_WIN
from pygls.lsp.types import DeleteFilesParams
from pygls.lsp.types import Diagnostic
from pygls.lsp.types import DiagnosticSeverity
from pygls.lsp.types import DidSaveTextDocumentParams
from pygls.lsp.types import InitializedParams
from pygls.lsp.types import InitializeParams
from pygls.lsp.types import MessageType
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from sphinx import __version__ as __sphinx_version__
from sphinx.application import Sphinx
from sphinx.domains import Domain
from sphinx.util import console
from sphinx.util.logging import SphinxLogRecord
from sphinx.util.logging import WarningLogRecordTranslator

from esbonio.cli import setup_cli
from esbonio.lsp.rst import LspHandler
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.rst import ServerConfig

try:
    from sphinx.util.logging import OnceFilter
except ImportError:
    # OnceFilter is not defined in Sphinx 2.x
    class OnceFilter:  # type: ignore
        def filter(self, *args, **kwargs):
            return True


# Order matters!
DEFAULT_MODULES = [
    "esbonio.lsp.directives",
    "esbonio.lsp.roles",
    "esbonio.lsp.sphinx.codeblocks",
    "esbonio.lsp.sphinx.domains",
    "esbonio.lsp.sphinx.images",
    "esbonio.lsp.sphinx.includes",
    "esbonio.lsp.sphinx.roles",
]
"""The modules to load in the default configuration of the server."""


DIAGNOSTIC_SEVERITY = {
    logging.ERROR: DiagnosticSeverity.Error,
    logging.INFO: DiagnosticSeverity.Information,
    logging.WARNING: DiagnosticSeverity.Warning,
}

PATH_VAR_PATTERN = re.compile(r"^\${(\w+)}/?.*")


class MissingConfigError(Exception):
    """Indicates that we couldn't locate the project's 'conf.py'"""


class SphinxConfig(BaseModel):
    """Used to represent either the current Sphinx configuration or the config options
    provided by the user at startup."""

    version: Optional[str]
    """Sphinx's version number."""

    conf_dir: Optional[str] = Field(None, alias="confDir")
    """Can be used to override the default conf.py discovery mechanism."""

    src_dir: Optional[str] = Field(None, alias="srcDir")
    """Can be used to override the default assumption on where the project's rst files
    are located."""

    build_dir: Optional[str] = Field(None, alias="buildDir")
    """Can be used to override the default location for storing build outputs."""

    builder_name: str = Field("html", alias="builderName")
    """The currently used builder name."""


class SphinxServerConfig(ServerConfig):

    hide_sphinx_output: bool = Field(False, alias="hideSphinxOutput")
    """A flag to indicate if Sphinx build output should be omitted from the log."""


class InitializationOptions(BaseModel):
    """The initialization options we can expect to receive from a client."""

    sphinx: SphinxConfig = Field(default_factory=SphinxConfig)
    """The ``esbonio.sphinx.*`` namespace of options."""

    server: SphinxServerConfig = Field(default_factory=SphinxServerConfig)
    """The ``esbonio.server.*`` namespace of options."""


class SphinxLogHandler(LspHandler):
    """A logging handler that can extract errors from Sphinx's build output."""

    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.app = app
        self.translator = WarningLogRecordTranslator(app)
        self.only_once = OnceFilter()
        self.diagnostics: Dict[str, List[Diagnostic]] = {}

    def get_location(self, location: str) -> Tuple[str, Optional[int]]:

        if not location:
            conf = pathlib.Path(self.app.confdir, "conf.py")
            return (str(conf), None)

        path, *parts = location.split(":")
        lineno = None

        # On windows the rest of the path will be the first element of parts
        if IS_WIN:
            path += f":{parts.pop(0)}"

        if len(parts) == 1:
            try:
                lineno = int(parts[0])
            except ValueError:
                pass

        if len(parts) == 2:
            # TODO: There's a possibility that there is an error in a docstring in a
            #       *.py file somewhere. In which case parts would look like
            #       ['docstring of {dotted.name}', '{lineno}']
            #
            #  e.g. ['docstring of esbonio.lsp.sphinx.SphinxLanguageServer.get_domains', '8']
            #
            #       It would be good to handle this case and look up the correct line
            #       number to place a diagnostic.
            pass

        return (Uri.from_fs_path(path), lineno)

    def emit(self, record: logging.LogRecord) -> None:

        conditions = [
            "sphinx" not in record.name,
            record.levelno not in {logging.WARNING, logging.ERROR},
            not self.translator,
        ]

        if any(conditions):
            # Log the record as normal
            super().emit(record)
            return

        # Only process errors/warnings once.
        if not self.only_once.filter(record):
            return

        # Let sphinx do what it does to warning/error messages
        self.translator.filter(record)

        loc = record.location if isinstance(record, SphinxLogRecord) else ""
        doc, lineno = self.get_location(loc)
        line = lineno or 1
        self.server.logger.debug("Reporting diagnostic at %s:%s", doc, line)

        try:
            message = record.msg % record.args
        except Exception:
            message = record.msg
            self.server.logger.debug(
                "Unable to format diagnostic message: %s", traceback.format_exc()
            )

        diagnostic = Diagnostic(
            range=Range(
                start=Position(line=line - 1, character=0),
                end=Position(line=line, character=0),
            ),
            message=message,
            severity=DIAGNOSTIC_SEVERITY.get(
                record.levelno, DiagnosticSeverity.Warning
            ),
        )

        if doc not in self.diagnostics:
            self.diagnostics[doc] = [diagnostic]
        else:
            self.diagnostics[doc].append(diagnostic)

        super().emit(record)


class SphinxLanguageServer(RstLanguageServer):
    """A language server dedicated to working with Sphinx projects."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.app: Optional[Sphinx] = None
        """The Sphinx application instance."""

        self._role_target_types: Optional[Dict[str, List[str]]] = None
        """Cache for role target types."""

        self._role_targets: Dict[str, List[tuple]] = {}
        """Cache for role target objects."""

    @property
    def configuration(self) -> Dict[str, Any]:
        """Return the server's actual configuration."""
        config = super().configuration

        if self.app:
            app = self.app
            builder_name = None if app.builder is None else app.builder.name
            config["sphinx"] = SphinxConfig(
                version=__sphinx_version__,
                confDir=app.confdir,
                srcDir=app.srcdir,
                buildDir=app.outdir,
                builderName=builder_name,
            )

        return config

    def initialize(self, params: InitializeParams):
        super().initialize(params)
        self.user_config = InitializationOptions(
            **typing.cast(Dict, params.initialization_options)
        )

    def initialized(self, params: InitializedParams):
        self.app = self._initialize_sphinx()
        self.build()

    def _initialize_sphinx(self):

        try:
            return self.create_sphinx_app(self.user_config)
        except MissingConfigError:
            self.show_message(
                message="Unable to find your 'conf.py', features that depend on Sphinx will be unavailable",
                msg_type=MessageType.Warning,
            )
            self.send_notification(
                "esbonio/buildComplete",
                {
                    "config": self.configuration,
                    "error": True,
                    "warnings": 0,
                },
            )
        except Exception:
            self.logger.error(traceback.format_exc())
            self.show_message(
                message="Unable to initialize Sphinx, see output window for details.",
                msg_type=MessageType.Error,
            )
            self.send_notification(
                "esbonio/buildComplete",
                {"config": self.configuration, "error": True, "warnings": 0},
            )

    def save(self, params: DidSaveTextDocumentParams):
        super().save(params)

        filepath = Uri.to_fs_path(params.text_document.uri)
        if filepath.endswith("conf.py"):
            if self.app:
                conf_dir = pathlib.Path(self.app.confdir)
            else:
                # The user's config is currently broken... where should their conf.py be?
                if self.user_config is not None:
                    config = typing.cast(InitializationOptions, self.user_config).sphinx
                else:
                    config = SphinxConfig()

                conf_dir = find_conf_dir(
                    self.workspace.root_uri, config
                ) or pathlib.Path(".")

            if str(conf_dir / "conf.py") == filepath:
                self.app = self._initialize_sphinx()
        else:
            self.clear_diagnostics("sphinx", params.text_document.uri)

        self.build()

    def delete_files(self, params: DeleteFilesParams):
        self.logger.debug("Deleted files: %s", params.files)

        # Files don't exist anymore, so diagnostics must be cleared.
        for file in params.files:
            self.clear_diagnostics("sphinx", file.uri)

        self.build()

    def build(self):

        if not self.app:
            return

        self.logger.debug("Building...")
        self.send_notification("esbonio/buildStart", {})

        # Reset the warnings counter
        self.app._warncount = 0
        error = False
        self.sphinx_log.diagnostics = {}

        try:
            self.app.build()
        except Exception:
            message = "Unable to build documentation, see output window for details."
            error = True

            self.logger.error(traceback.format_exc())
            self.show_message(message=message, msg_type=MessageType.Error)

        for doc, diagnostics in self.sphinx_log.diagnostics.items():
            self.logger.debug("Found %d problems for %s", len(diagnostics), doc)
            self.set_diagnostics("sphinx", doc, diagnostics)

        self.sync_diagnostics()
        self.send_notification(
            "esbonio/buildComplete",
            {
                "config": self.configuration,
                "error": error,
                "warnings": self.app._warncount,
            },
        )

    def create_sphinx_app(self, options: InitializationOptions) -> Optional[Sphinx]:
        """Create a Sphinx application instance with the given config."""
        sphinx = options.sphinx
        server = options.server

        self.logger.debug("Workspace root %s", self.workspace.root_uri)
        self.logger.debug("Sphinx Config %s", sphinx.dict())

        conf_dir = find_conf_dir(self.workspace.root_uri, sphinx)
        if conf_dir is None:
            raise MissingConfigError()

        builder_name = sphinx.builder_name
        src_dir = get_src_dir(self.workspace.root_uri, conf_dir, sphinx)
        build_dir = get_build_dir(self.workspace.root_uri, conf_dir, sphinx)
        doctree_dir = build_dir / "doctrees"
        build_dir /= builder_name

        self.logger.debug("Config dir %s", conf_dir)
        self.logger.debug("Src dir %s", src_dir)
        self.logger.debug("Build dir %s", build_dir)
        self.logger.debug("Doctree dir %s", doctree_dir)

        # Disable color escape codes in Sphinx's log messages
        console.nocolor()

        app = Sphinx(
            srcdir=str(src_dir),
            confdir=str(conf_dir),
            outdir=str(build_dir),
            doctreedir=str(doctree_dir),
            buildername=builder_name,
            status=None,  # type: ignore
            warning=None,  # type: ignore
            freshenv=True,  # Have Sphinx reload everything on first build.
        )

        # This has to happen after app creation otherwise our handler
        # will get cleared by Sphinx's setup.
        if not server.hide_sphinx_output:
            sphinx_logger = logging.getLogger("sphinx")
            sphinx_logger.setLevel(logging.INFO)

            self.sphinx_log = SphinxLogHandler(app, self)
            self.sphinx_log.setLevel(logging.INFO)

            formatter = logging.Formatter("%(message)s")
            self.sphinx_log.setFormatter(formatter)
            sphinx_logger.addHandler(self.sphinx_log)

        return app

    def get_doctree(
        self, *, docname: Optional[str] = None, uri: Optional[str] = None
    ) -> Optional[Any]:
        """Return the doctree that corresponds with the specified document.

        The ``docname`` of a document is its path relative to the project's ``srcdir``
        minus the extension e.g. the docname of the file ``docs/lsp/features.rst``
        would be ``lsp/features``.

        Parameters
        ----------
        docname:
           Returns the doctree that corresponds with the given docname
        uri:
           Returns the doctree that corresponds with the given uri.
        """

        if self.app is None or self.app.env is None or self.app.builder is None:
            return None

        if uri is not None:
            fspath = Uri.to_fs_path(uri)
            docname = self.app.env.path2doc(fspath)

        if docname is None:
            return None

        try:
            return self.app.env.get_and_resolve_doctree(docname, self.app.builder)
        except FileNotFoundError:
            self.logger.debug(traceback.format_exc())
            return None

    def get_domain(self, name: str) -> Optional[Domain]:
        """Return the domain with the given name.

        If a domain with the given name cannot be found, this method will return None.

        Parameters
        ----------
        name:
           The name of the domain
        """

        if self.app is None or self.app.env is None:
            return None

        domains = self.app.env.domains
        return domains.get(name, None)

    def get_domains(self) -> Iterator[Tuple[str, Domain]]:
        """Get all the domains registered with an applications.

        Returns a generator that iterates through all of an application's domains,
        taking into account configuration variables such as ``primary_domain``.
        Yielded values will be a tuple of the form ``(prefix, domain)`` where

        - ``prefix`` is the namespace that should be used when referencing items
          in the domain
        - ``domain`` is the domain object itself.

        """

        if self.app is None or self.app.env is None:
            return []

        domains = self.app.env.domains
        primary_domain = self.app.config.primary_domain

        for name, domain in domains.items():
            prefix = name

            # Items from the standard and primary domains don't require the namespace prefix
            if name == "std" or name == primary_domain:
                prefix = ""

            yield prefix, domain

    def get_directives(self) -> Dict[str, Directive]:
        """Return a dictionary of the known directives"""

        if self._directives is not None:
            return self._directives

        self._directives = super().get_directives()

        for prefix, domain in self.get_domains():
            fmt = "{prefix}:{name}" if prefix else "{name}"

            for name, directive in domain.directives.items():
                key = fmt.format(name=name, prefix=prefix)
                self._directives[key] = directive

        return self._directives

    def get_directive_options(self, name: str) -> Dict[str, Any]:
        """Return the options specification for the given directive."""

        directive = self.get_directives().get(name, None)
        if directive is None:
            return {}

        options = directive.option_spec

        if name.startswith("auto") and self.app:
            self.logger.debug("Processing options for '%s' directive", name)
            name = name.replace("auto", "")

            self.logger.debug("Documenter name is '%s'", name)
            documenter = self.app.registry.documenters.get(name, None)

            if documenter is not None:
                options = documenter.option_spec

        return options or {}

    def get_roles(self) -> Dict[str, Any]:
        """Return a dictionary of known roles."""

        if self._roles is not None:
            return self._roles

        self._roles = super().get_roles()

        for prefix, domain in self.get_domains():
            fmt = "{prefix}:{name}" if prefix else "{name}"

            for name, role in domain.roles.items():
                key = fmt.format(name=name, prefix=prefix)
                self._roles[key] = role

        return self._roles

    def get_default_role(self) -> Tuple[Optional[str], Optional[str]]:
        """Return the project's default role"""

        if not self.app:
            return None, None

        role = self.app.config.default_role
        if not role:
            return None, None

        if ":" in role:
            domain, name = role.split(":")

            if domain == self.app.config.primary_domain:
                domain = ""

            return domain, name

        return None, role

    def get_role_target_types(self, name: str, domain_name: str = "") -> List[str]:
        """Return a map indicating which object types a role is capable of linking
        with.

        For example

        .. code-block:: python

           {
               "func": ["py:function"],
               "class": ["py:class", "py:exception"]
           }
        """

        key = f"{domain_name}:{name}" if domain_name else name

        if self._role_target_types is not None:
            return self._role_target_types.get(key, [])

        self._role_target_types = {}

        for prefix, domain in self.get_domains():
            fmt = "{prefix}:{name}" if prefix else "{name}"

            for name, item_type in domain.object_types.items():
                for role in item_type.roles:
                    role_key = fmt.format(name=role, prefix=prefix)
                    target_types = self._role_target_types.get(role_key, list())
                    target_types.append(fmt.format(name=name, prefix=prefix))

                    self._role_target_types[role_key] = target_types

        types = self._role_target_types.get(key, [])
        self.logger.debug("Role '%s' targets object types '%s'", key, types)

        return types

    def get_role_targets(self, name: str, domain: str = "") -> List[tuple]:
        """Return a list of objects targeted by the given role.

        Parameters
        ----------
        name:
           The name of the role
        domain:
           The domain the role is a part of, if applicable.
        """

        if not self._role_targets:
            self._index_role_targets()

        targets: List[tuple] = []
        for target_type in self.get_role_target_types(name, domain):
            targets += self._role_targets.get(target_type, [])

        return targets

    def _index_role_targets(self):
        self._role_targets = {}

        for prefix, domain_obj in self.get_domains():
            fmt = "{prefix}:{name}" if prefix else "{name}"

            for obj in domain_obj.get_objects():
                obj_key = fmt.format(name=obj[2], prefix=prefix)
                objects = self._role_targets.get(obj_key, list())
                objects.append(obj)

                self._role_targets[obj_key] = objects

    def get_intersphinx_projects(self) -> List[str]:
        """Return the list of configured intersphinx project names."""

        if self.app is None:
            return []

        inv = getattr(self.app.env, "intersphinx_named_inventory", {})
        return list(inv.keys())

    def has_intersphinx_targets(
        self, project: str, name: str, domain: str = ""
    ) -> bool:
        """Return ``True`` if the given intersphinx project has targets targeted by the
        given role.

        Parameters
        ----------
        project:
           The project to check
        name:
           The name of the role
        domain:
           The domain the role is a part of, if applicable.
        """

        targets = self.get_intersphinx_targets(project, name, domain)

        if len(targets) == 0:
            return False

        return any([len(items) > 0 for items in targets.values()])

    def get_intersphinx_targets(
        self, project: str, name: str, domain: str = ""
    ) -> Dict[str, Dict[str, tuple]]:
        """Return the intersphinx objects targeted by the given role.

        Parameters
        ----------
        project:
           The project to return targets from
        name:
           The name of the role
        domain:
           The domain the role is a part of, if applicable.
        """

        if self.app is None:
            return {}

        inv = getattr(self.app.env, "intersphinx_named_inventory", {})
        if project not in inv:
            return {}

        targets = {}
        inv = inv[project]
        primary_domain = self.app.config.primary_domain or ""

        for target_type in self.get_role_target_types(name, domain):

            if target_type in inv:
                targets[target_type] = inv[target_type]
                continue

            # Intersphinx targets are always namespaced, so we would need to be explicit
            # about the domain the type sits in.
            if f"{primary_domain}:{target_type}" in inv:
                targets[target_type] = inv[f"{primary_domain}:{target_type}"]
                continue

            # The 'std' domain must also be considered.
            if f"std:{target_type}" in inv:
                targets[target_type] = inv[f"std:{target_type}"]

        return targets


def find_conf_dir(root_uri: str, config: SphinxConfig) -> Optional[pathlib.Path]:
    """Attempt to find Sphinx's configuration file within the given workspace."""

    root = Uri.to_fs_path(root_uri)

    if config.conf_dir:
        return expand_conf_dir(root, config.conf_dir)

    ignore_paths = [".tox", "site-packages"]

    for candidate in pathlib.Path(root).glob("**/conf.py"):
        # Skip any files that obviously aren't part of the project
        if any(path in str(candidate) for path in ignore_paths):
            continue

        return candidate.parent

    return None


def expand_conf_dir(root_dir: str, conf_dir: str) -> pathlib.Path:
    """Expand the user provided conf_dir into a real path.

    Here is where we handle "variables" that can be included in the path, currently
    we support

    - ``${workspaceRoot}`` which expands to the workspace root as provided by the
      language client.

    Parameters
    ----------
    root_dir:
       The workspace root path
    conf_dir:
       The user provided path
    """

    match = PATH_VAR_PATTERN.match(conf_dir)
    if not match or match.group(1) != "workspaceRoot":
        return pathlib.Path(conf_dir)

    conf = pathlib.Path(conf_dir).parts[1:]
    return pathlib.Path(root_dir, *conf).resolve()


def get_src_dir(
    root_uri: str, conf_dir: pathlib.Path, config: SphinxConfig
) -> pathlib.Path:
    """Get the src dir to use based on the given conifg.

    By default the src dir will be the same as the conf dir, but this can
    be overriden by the given conifg.

    There are a number of "variables" that can be included in the path,
    currently we support

    - ``${workspaceRoot}`` which expands to the workspace root as provided
      by the language client.
    - ``${confDir}`` which expands to the configured config dir.

    Parameters
    ----------
    root_uri:
       The workspace root uri
    conf_dir:
       The project's conf dir
    config:
       The user's configuration.
    """

    if not config.src_dir:
        return conf_dir

    src_dir = config.src_dir
    root_dir = Uri.to_fs_path(root_uri)

    match = PATH_VAR_PATTERN.match(src_dir)
    if match and match.group(1) == "workspaceRoot":
        src = pathlib.Path(src_dir).parts[1:]
        return pathlib.Path(root_dir, *src).resolve()

    if match and match.group(1) == "confDir":
        src = pathlib.Path(src_dir).parts[1:]
        return pathlib.Path(conf_dir, *src).resolve()

    return pathlib.Path(src_dir)


def get_build_dir(
    root_uri: str, conf_dir: pathlib.Path, config: SphinxConfig
) -> pathlib.Path:
    """Get the build dir to use based on the given conifg.

    If nothing is specified in the given ``config``, this will choose a location within
    the user's cache dir (as determined by `appdirs <https://pypi.org/project/appdirs>`).
    The directory name will be a hash derived from the given ``conf_dir`` for the
    project.

    Alternatively the user (or least language client) can override this by setting
    either an absolute path, or a path based on the following "variables".

    - ``${workspaceRoot}`` which expands to the workspace root as provided
      by the language client.
    - ``${confDir}`` which expands to the configured config dir.

    Parameters
    ----------
    root_uri:
       The workspace root uri
    conf_dir:
       The project's conf dir
    config:
       The user's configuration.
    """

    if not config.build_dir:
        # Try to pick a sensible dir based on the project's location
        cache = appdirs.user_cache_dir("esbonio", "swyddfa")
        project = hashlib.md5(str(conf_dir).encode()).hexdigest()

        return pathlib.Path(cache) / project

    root_dir = Uri.to_fs_path(root_uri)
    match = PATH_VAR_PATTERN.match(config.build_dir)

    if match and match.group(1) == "workspaceRoot":
        build = pathlib.Path(config.build_dir).parts[1:]
        return pathlib.Path(root_dir, *build).resolve()

    if match and match.group(1) == "confDir":
        build = pathlib.Path(config.build_dir).parts[1:]
        return pathlib.Path(conf_dir, *build).resolve()

    # Convert path to/from uri so that any path quirks from windows are
    # automatically handled
    build_uri = Uri.from_fs_path(config.build_dir)
    build_dir = Uri.to_fs_path(build_uri)

    return pathlib.Path(build_dir)


cli = setup_cli("esbonio.lsp.sphinx", "Esbonio's Sphinx language server.")
cli.set_defaults(modules=DEFAULT_MODULES)
cli.set_defaults(server_cls=SphinxLanguageServer)



"""Role support."""
import json
import re
import typing
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import pkg_resources
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionItemKind
from pygls.lsp.types import Location
from pygls.lsp.types import MarkupContent
from pygls.lsp.types import MarkupKind
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from pygls.lsp.types import TextEdit

from esbonio.lsp.directives import DIRECTIVE
from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.rst import DefinitionContext
from esbonio.lsp.rst import LanguageFeature
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.sphinx import SphinxLanguageServer

try:
    from typing import Protocol
except ImportError:
    # Protocol is only available in Python 3.8+
    class Protocol:  # type: ignore
        ...


ROLE = re.compile(
    r"""
    ([^\w:]|^\s*)                     # roles cannot be preceeded by letter chars
    (?P<role>
      :                               # roles begin with a ':' character
      (?!:)                           # the next character cannot be a ':'
      ((?P<domain>[\w]+):(?=\w))?     # roles may include a domain (that must be followed by a word character)
      ((?P<name>[\w-]+):?)?           # roles have a name
    )
    (?P<target>
      `                               # targets begin with a '`' character
      ((?P<alias>[^<`>]*?)<)?         # targets may specify an alias
      (?P<modifier>[!~])?             # targets may have a modifier
      (?P<label>[^<`>]*)?             # targets contain a label
      >?                              # labels end with a '>' when there's an alias
      `?                              # targets end with a '`' character
    )?
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse parial and complete roles.

I'm not sure if there are offical names for the components of a role, but the
language server breaks a role down into a number of parts::

                 vvvvvv label
                v modifier(optional)
               vvvvvvvv target
   :c:function:`!malloc`
   ^^^^^^^^^^^^ role
      ^^^^^^^^ name
    ^ domain (optional)

The language server sometimes refers to the above as a "plain" role, in that the
role's target contains just the label of the object it is linking to. However it's
also possible to define "aliased" roles, where the link text in the final document
is overriden, for example::

                vvvvvvvvvvvvvvvvvvvvvvvv alias
                                          vvvvvv label
                                         v modifier (optional)
               vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv target
   :c:function:`used to allocate memory <~malloc>`
   ^^^^^^^^^^^^ role
      ^^^^^^^^ name
    ^ domain (optional)

See :func:`tests.test_roles.test_role_regex` for a list of example strings this pattern
is expected to match.
"""


DEFAULT_ROLE = re.compile(
    r"""
    (?<![:`])
    (?P<target>
      `                               # targets begin with a '`' character
      ((?P<alias>[^<`>]*?)<)?         # targets may specify an alias
      (?P<modifier>[!~])?             # targets may have a modifier
      (?P<label>[^<`>]*)?             # targets contain a label
      >?                              # labels end with a '>' when there's an alias
      `?                              # targets end with a '`' character
    )
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse parial and complete "default" roles.

A "default" role is the target part of a normal role - but without the ``:name:`` part.
"""


class TargetDefinition(Protocol):
    """A definition provider for role targets"""

    def find_definitions(
        self, context: DefinitionContext, name: str, domain: Optional[str]
    ) -> List[Location]:
        """Return a list of locations representing the definition of the given role
        target.

        Parameters
        ----------
        doc:
           The document containing the match
        match:
           The match object that triggered the definition request
        name:
           The name of the role
        domain:
           The domain the role is part of, if applicable.
        """


class TargetCompletion(Protocol):
    """A completion provider for role targets"""

    def complete_targets(
        self, context: CompletionContext, name: str, domain: Optional[str]
    ) -> List[CompletionItem]:
        """Return a list of completion items representing valid targets for the given
        role.

        Parameters
        ----------
        context:
           The completion context
        domain:
           The name of the domain the role is a member of
        name:
           The name of the role to generate completion suggestions for.
        """


class Roles(LanguageFeature):
    """Role support for the language server."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._documentation: Dict[str, Dict[str, str]] = {}
        """Cache for documentation."""

        self._target_definition_providers: List[TargetDefinition] = []
        """A list of providers that locate the definition for the given role target."""

        self._target_completion_providers: List[TargetCompletion] = []
        """A list of providers that give completion suggestions for role target
        objects."""

    def add_target_definition_provider(self, provider: TargetDefinition) -> None:
        self._target_definition_providers.append(provider)

    def add_target_completion_provider(self, provider: TargetCompletion) -> None:
        self._target_completion_providers.append(provider)

    def add_documentation(self, documentation: Dict[str, Dict[str, Any]]) -> None:
        """Register role documentation.

        ``documentation`` should be a dictionary of the form ::

           documentation = {
               "raw(docutils.parsers.rst.roles.raw_role)": {
                   "is_markdown": true,
                   "license": "https://...",
                   "source": "https://...",
                   "description": [
                       "# :raw:",
                       "The raw role is used for...",
                       ...
                   ]
               }
           }

        where the key is of the form `name(dotted_name)`. There are cases where a role's
        implementation is not sufficient to uniquely identify it as multiple roles can
        be provided by a single class.

        This means the key has to be a combination of the ``name`` the user writes in
        an reStructuredText document and ``dotted_name`` is the fully qualified name of
        the role's implementation.

        .. note::

           If there is a clash with an existing key, the existing value will be
           overwritten with the new value.

        The values in this dictionary are themselves dictionaries with the following
        fields.

        ``description``
           A list of strings for the role's usage.

        ``is_markdown``
           A boolean flag used to indicate whether the ``description`` is written in
           plain text or markdown.

        ``source``
           The url to the documentation's source.

        ``license``
           The url to the documentation's license.

        Parameters
        ----------
        documentation:
           The documentation to register.
        """

        for key, doc in documentation.items():
            description = doc.get("description", [])
            if not description:
                continue

            source = doc.get("source", "")
            if source:
                description.append(f"\n[Source]({source})")

            license = doc.get("license", "")
            if license:
                description.append(f"\n[License]({license})")

            doc["description"] = "\n".join(description)
            self._documentation[key] = doc

    completion_triggers = [ROLE, DEFAULT_ROLE]
    definition_triggers = [ROLE]

    def definition(self, context: DefinitionContext) -> List[Location]:

        domain = context.match.group("domain") or None
        name = context.match.group("name")

        definitions = []

        for provide in self._target_definition_providers:
            definitions += provide.find_definitions(context, name, domain) or []

        return definitions

    def complete(self, context: CompletionContext) -> List[CompletionItem]:
        """Generate completion suggestions relevant to the current context.

        This function is a little intense, but its sole purpose is to determine the
        context in which the completion request is being made and either return
        nothing, or the results of :meth:`~esbonio.lsp.roles.Roles.complete_roles` or
        :meth:`esbonio.lsp.roles.Roles.complete_targets` whichever is appropriate.

        Parameters
        ----------
        context:
           The context of the completion request.
        """

        # Do not suggest completions within the middle of Python code.
        if context.location == "py":
            return []

        groups = context.match.groupdict()
        target = groups["target"]

        # All text matched by the regex
        text = context.match.group(0)
        start, end = context.match.span()

        if target:
            target_index = start + text.find(target)

            # Only trigger target completions if the request was made from within
            # the target part of the role.
            if target_index <= context.position.character <= end:
                return self.complete_targets(context)

        # If there's no indent, then this can only be a role definition
        indent = context.match.group(1)
        if indent == "":
            return self.complete_roles(context)

        # Otherwise, search backwards until we find a blank line or an unindent
        # so that we can determine the appropriate context.
        linum = context.position.line - 1

        try:
            line = context.doc.lines[linum]
        except IndexError:
            return self.complete_roles(context)

        while linum >= 0 and line.startswith(indent):
            linum -= 1
            line = context.doc.lines[linum]

        # Unless we are within a directive's options block, we should offer role
        # suggestions
        if DIRECTIVE.match(line):
            return []

        return self.complete_roles(context)

    def completion_resolve(self, item: CompletionItem) -> CompletionItem:

        # We need extra info to know who to call
        if not item.data:
            return item

        data = typing.cast(Dict, item.data)
        ctype = data.get("completion_type", "")

        if ctype == "role":
            return self.completion_resolve_role(item)

        return item

    def complete_roles(self, context: CompletionContext) -> List[CompletionItem]:

        match = context.match
        groups = match.groupdict()
        domain = groups["domain"] or ""
        items = []

        # Insert text starting from the starting ':' character of the role.
        start = match.span()[0] + match.group(0).find(":")
        end = start + len(groups["role"])

        range_ = Range(
            start=Position(line=context.position.line, character=start),
            end=Position(line=context.position.line, character=end),
        )

        for name, role in self.rst.get_roles().items():

            if not name.startswith(domain):
                continue

            try:
                dotted_name = f"{role.__module__}.{role.__name__}"
            except AttributeError:
                dotted_name = f"{role.__module__}.{role.__class__.__name__}"

            insert_text = f":{name}:"
            item = CompletionItem(
                label=name,
                kind=CompletionItemKind.Function,
                detail=f"{dotted_name}",
                filter_text=insert_text,
                text_edit=TextEdit(range=range_, new_text=insert_text),
                data={"completion_type": "role"},
            )

            items.append(item)

        return items

    def completion_resolve_role(self, item: CompletionItem) -> CompletionItem:

        # We need the detail field set to the role implementation's fully qualified name
        if not item.detail:
            return item

        documentation = self.get_documentation(item.label, item.detail)
        if not documentation:
            return item

        description = documentation.get("description", "")
        is_markdown = documentation.get("is_markdown", False)
        kind = MarkupKind.Markdown if is_markdown else MarkupKind.PlainText

        item.documentation = MarkupContent(kind=kind, value=description)
        return item

    def complete_targets(self, context: CompletionContext) -> List[CompletionItem]:
        """Generate the list of role target completion suggestions."""

        groups = context.match.groupdict()

        # Handle the default role case.
        if "role" not in groups:
            domain, name = self.rst.get_default_role()
            if not name:
                return []
        else:
            name = groups["name"]
            domain = groups["domain"]

        domain = domain or ""
        name = name or ""

        # Only generate suggestions for "aliased" targets if the request comes from
        # within the <> chars.
        if groups["alias"]:
            text = context.match.group(0)
            start = context.match.span()[0] + text.find(groups["alias"])
            end = start + len(groups["alias"])

            if start <= context.position.character <= end:
                return []

        targets = []

        startchar = "<" if "<" in groups["target"] else "`"
        endchars = ">`" if "<" in groups["target"] else "`"

        start, end = context.match.span()
        start += context.match.group(0).index(startchar) + 1
        range_ = Range(
            start=Position(line=context.position.line, character=start),
            end=Position(line=context.position.line, character=end),
        )
        prefix = context.match.group(0)[start:]
        modifier = groups["modifier"] or ""

        for provide in self._target_completion_providers:
            candidates = provide.complete_targets(context, name, domain) or []

            for candidate in candidates:

                # Don't interfere with items that already carry a `text_edit`, allowing
                # some providers (like filepaths) to do something special.
                if not candidate.text_edit:
                    new_text = candidate.insert_text or candidate.label

                    # This is rather annoying, but `filter_text` needs to start with
                    # the text we are going to replace, otherwise VSCode won't show our
                    # suggestions!
                    candidate.filter_text = f"{prefix}{new_text}"

                    candidate.text_edit = TextEdit(
                        range=range_, new_text=f"{modifier}{new_text}"
                    )
                    candidate.insert_text = None

                if not candidate.text_edit.new_text.endswith(endchars):
                    candidate.text_edit.new_text += endchars

                targets.append(candidate)

        return targets

    def get_documentation(
        self, label: str, implementation: str
    ) -> Optional[Dict[str, Any]]:
        """Return the documentation for the given role, if available.

        If documentation for the given ``label`` cannot be found, this function will also
        look for the label under the project's :confval:`sphinx:primary_domain` followed
        by the ``std`` domain.

        Parameters
        ----------
        label:
           The name of the role, as the user would type in an reStructuredText file.
        implementation:
           The full dotted name of the role's implementation.
        """

        key = f"{label}({implementation})"
        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        if not isinstance(self.rst, SphinxLanguageServer) or not self.rst.app:
            return None

        # Nothing found, try the primary domain
        domain = self.rst.app.config.primary_domain
        key = f"{domain}:{label}({implementation})"

        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        # Still nothing, try the standard domain
        key = f"std:{label}({implementation})"

        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        return None


def esbonio_setup(rst: RstLanguageServer):
    roles = Roles(rst)
    rst.add_feature(roles)

    docutils_docs = pkg_resources.resource_string("esbonio.lsp.rst", "roles.json")
    roles.add_documentation(json.loads(docutils_docs.decode("utf8")))

    if isinstance(rst, SphinxLanguageServer):
        sphinx_docs = pkg_resources.resource_string("esbonio.lsp.sphinx", "roles.json")
        roles.add_documentation(json.loads(sphinx_docs.decode("utf8")))



import os.path
import typing
from typing import List

import pygls.uris as Uri
from pygls.lsp.types import CompletionItem

from esbonio.lsp.directives import Directives
from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.sphinx import SphinxLanguageServer
from esbonio.lsp.util.filepaths import complete_sphinx_filepaths
from esbonio.lsp.util.filepaths import path_to_completion_item


class Images:
    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)

    def complete_arguments(
        self, context: CompletionContext, domain: str, name: str
    ) -> List[CompletionItem]:

        if domain or name not in {"figure", "image"}:
            return []

        if not self.rst.app:
            return []

        srcdir = self.rst.app.srcdir
        partial = context.match.group("argument")
        base = os.path.dirname(Uri.to_fs_path(context.doc.uri))
        items = complete_sphinx_filepaths(srcdir, base, partial)

        return [path_to_completion_item(context, p) for p in items]


def esbonio_setup(rst: RstLanguageServer):

    name = "esbonio.lsp.directives.Directives"
    directives = rst.get_feature(name)

    if not isinstance(rst, SphinxLanguageServer) or not directives:
        return

    # To keep mypy happy.
    directives = typing.cast(Directives, directives)

    directives.add_argument_completion_provider(Images(rst))



"""Logic around directive completions goes here."""
import json
import re
import typing
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import pkg_resources
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionItemKind
from pygls.lsp.types import InsertTextFormat
from pygls.lsp.types import Location
from pygls.lsp.types import MarkupContent
from pygls.lsp.types import MarkupKind
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from pygls.lsp.types import TextEdit

from esbonio.lsp import CompletionContext
from esbonio.lsp import DefinitionContext
from esbonio.lsp import LanguageFeature
from esbonio.lsp import RstLanguageServer
from esbonio.lsp.sphinx import SphinxLanguageServer

try:
    from typing import Protocol
except ImportError:
    # Protocol is only available in Python 3.8+
    class Protocol:  # type: ignore
        ...


DIRECTIVE = re.compile(
    r"""
    (\s*)                           # directives can be indented
    (?P<directive>
      \.\.                          # directives start with a comment
      [ ]?                          # followed by a space
      ((?P<domain>[\w]+):(?!:))?    # directives may include a domain
      (?P<name>([\w-]|:(?!:))+)?    # directives have a name
      (::)?                         # directives end with '::'
    )
    ([\s]+(?P<argument>.*$))?       # directives may take an argument
    """,
    re.VERBOSE,
)
"""A regular expression to detect and parse partial and complete directives.

This does **not** include any options or content that may be included underneath
the initial declaration. The language server breaks a directive down into a number
of parts::

                   vvvvvv argument
   .. c:function:: malloc
   ^^^^^^^^^^^^^^^ directive
        ^^^^^^^^ name
      ^ domain (optional)
"""


DIRECTIVE_OPTION = re.compile(
    r"""
    (?P<indent>\s+)       # directive options must be indented
    (?P<option>
      :                   # options start with a ':'
      (?P<name>[\w-]+)?   # options have a name
      :?                  # options end with a ':'
    )
    (\s*
      (?P<value>.*)       # options can have a value
    )?
    """,
    re.VERBOSE,
)
"""A regular expression used to detect and parse partial and complete directive options.

The language server breaks an option down into a number of parts::

               vvvvvv value
   |   :align: center
       ^^^^^^^ option
        ^^^^^ name
    ^^^ indent
"""


class ArgumentCompletion(Protocol):
    """A completion provider for directive arguments."""

    def complete_arguments(
        self, context: CompletionContext, domain: str, name: str
    ) -> List[CompletionItem]:
        """Return a list of completion items representing valid targets for the given
        directive.

        Parameters
        ----------
        context:
           The completion context
        domain:
           The name of the domain the directive is a member of
        name:
           The name of the domain
        """


class ArgumentDefinition(Protocol):
    """A definition provider for directive arguments."""

    def find_definitions(
        self,
        context: DefinitionContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> List[Location]:
        """Return a list of locations representing definitions of the given argument.

        Parameters
        ----------
        context:
           The context of the definition request.
        directive:
           The name of the directive the argument is associated with.
        domain:
           The name of the domain the directive belongs to, if applicable.
        argument:
           The argument to find the definition of.
        """


class Directives(LanguageFeature):
    """Directive support for the language server."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._documentation: Dict[str, Dict[str, str]] = {}
        """Cache for documentation."""

        self._argument_completion_providers: Dict[str, ArgumentCompletion] = {}
        """A dictionary of providers that give completion suggestions for directive
        arguments."""

        self._argument_definition_providers: Dict[str, ArgumentDefinition] = {}
        """A dictionary of providers that give completion suggestions for directive
        arguments."""

    def add_argument_completion_provider(self, provider: ArgumentCompletion) -> None:
        """Register an :class:`~esbonio.lsp.directives.ArgumentCompletion` provider.

        Parameters
        ----------
        provider:
           The provider to register.
        """
        key = f"{provider.__module__}.{provider.__class__.__name__}"
        self._argument_completion_providers[key] = provider

    def add_argument_definition_provider(self, provider: ArgumentDefinition) -> None:
        """Register an :class:`~esbonio.lsp.directives.ArgumentCompletion` provider.

        Parameters
        ----------
        provider:
           The provider to register.
        """
        key = f"{provider.__module__}.{provider.__class__.__name__}"
        self._argument_definition_providers[key] = provider

    def add_documentation(self, documentation: Dict[str, Dict[str, Any]]) -> None:
        """Register directive documentation.

        ``documentation`` should be a dictionary with the following structure ::

           documentation = {
               "raw(docutils.parsers.rst.directives.misc.Raw)": {
                   "is_markdown": true,
                   "license": "https://...",
                   "source": "https://...",
                   "description": [
                       "# .. raw::",
                       "The raw directive is used for...",
                       ...
                    ]
                   "options": {
                       "file": "The file option allows...",
                       ...
                   }
               }
           }

        where the key has the form ``name(dotted_name)``. There are cases where a
        directive's implementation is not sufficient to uniquely identify it as
        multiple directives can be provided by a single class.

        This means the key has to be a combination of the ``name`` the user writes
        in a reStructuredText document and ``dotted_name`` is the fully qualified
        class name of the directive's implementation.

        .. note::

           If there is a clash with an existing key, the existing value will be
           overwritten with the new value.

        The values in this dictionary are themselves dictionaries with the following
        fields.

        ``description``
           A list of strings for the directive's main documentation.

        ``options``,
           A dictionary, with a field for the documentaton of each of the directive's
           options.

        ``is_markdown``
           A boolean flag used to indicate whether the ``description`` and ``options``
           are written in plain text or markdown.

        ``source``
           The url to the documentation's source

        ``license``
           The url to the documentation's license

        Parameters
        ----------
        documentation:
           The documentation to register.
        """

        for key, doc in documentation.items():
            description = doc.get("description", [])

            if not description:
                continue

            source = doc.get("source", "")
            if source:
                description.append(f"\n[Source]({source})")

            license = doc.get("license", "")
            if license:
                description.append(f"\n[License]({license})")

            doc["description"] = "\n".join(description)
            self._documentation[key] = doc

    completion_triggers = [DIRECTIVE, DIRECTIVE_OPTION]

    def completion_resolve(self, item: CompletionItem) -> CompletionItem:

        # We need extra info to know who to call.
        if not item.data:
            return item

        data = typing.cast(Dict, item.data)
        ctype = data.get("completion_type", "")

        if ctype == "directive":
            return self.completion_resolve_directive(item)

        if ctype == "directive_option":
            return self.completion_resolve_option(item)

        return item

    def complete(self, context: CompletionContext) -> List[CompletionItem]:

        # Do not suggest completions within the middle of Python code.
        if context.location == "py":
            return []

        groups = context.match.groupdict()

        # Are we completing a directive's options?
        if "directive" not in groups:
            return self.complete_options(context)

        # Are we completing the directive's argument?
        directive_end = context.match.span()[0] + len(groups["directive"])
        complete_directive = groups["directive"].endswith("::")

        if complete_directive and directive_end < context.position.character:
            return self.complete_arguments(context)

        return self.complete_directives(context)

    def complete_arguments(self, context: CompletionContext) -> List[CompletionItem]:
        arguments = []
        name = context.match.group("name")
        domain = context.match.group("domain") or ""

        for provider_name, provider in self._argument_completion_providers.items():
            arguments += provider.complete_arguments(context, domain, name) or []

        return arguments

    def complete_directives(self, context: CompletionContext) -> List[CompletionItem]:
        self.logger.debug("Completing directives")

        items = []
        match = context.match
        groups = match.groupdict()

        domain = ""
        if groups["domain"]:
            domain = f'{groups["domain"]}:'

        # Calculate the range of text the CompletionItems should edit.
        # If there is an existing argument to the directive, we should leave it untouched
        # otherwise, edit the whole line to insert any required arguments.
        start = match.span()[0] + match.group(0).find(".")
        include_argument = True
        end = match.span()[1]

        if groups["argument"]:
            include_argument = False
            end = match.span()[0] + match.group(0).find("::") + 2

        range_ = Range(
            start=Position(line=context.position.line, character=start),
            end=Position(line=context.position.line, character=end),
        )

        for name, directive in self.rst.get_directives().items():

            if not name.startswith(domain):
                continue

            # TODO: Give better names to arguments based on what they represent.
            if include_argument:
                insert_format = InsertTextFormat.Snippet
                args = " " + " ".join(
                    "${{{0}:arg{0}}}".format(i)
                    for i in range(1, directive.required_arguments + 1)
                )
            else:
                args = ""
                insert_format = InsertTextFormat.PlainText

            insert_text = f".. {name}::{args}"

            items.append(
                CompletionItem(
                    label=name,
                    kind=CompletionItemKind.Class,
                    detail=f"{directive.__module__}.{directive.__name__}",
                    filter_text=insert_text,
                    text_edit=TextEdit(range=range_, new_text=insert_text),
                    insert_text_format=insert_format,
                    data={"completion_type": "directive"},
                )
            )

        return items

    def completion_resolve_directive(self, item: CompletionItem) -> CompletionItem:

        # We need the detail field set to the implementation's fully qualified name.
        if not item.detail:
            return item

        documentation = self.get_documentation(item.label, item.detail)
        if not documentation:
            return item

        description = documentation.get("description", "")
        is_markdown = documentation.get("is_markdown", False)
        kind = MarkupKind.Markdown if is_markdown else MarkupKind.PlainText

        item.documentation = MarkupContent(kind=kind, value=description)
        return item

    def complete_options(self, context: CompletionContext) -> List[CompletionItem]:

        surrounding_directive = self.get_surrounding_directive(context)
        if not surrounding_directive:
            return []

        domain = ""
        if surrounding_directive.group("domain"):
            domain = f'{surrounding_directive.group("domain")}:'

        name = f"{domain}{surrounding_directive.group('name')}"
        directive = self.rst.get_directives().get(name, None)

        if not directive:
            return []

        items = []
        match = context.match
        groups = match.groupdict()

        option = groups["option"]
        start = match.span()[0] + match.group(0).find(option)
        end = start + len(option)

        range_ = Range(
            start=Position(line=context.position.line, character=start),
            end=Position(line=context.position.line, character=end),
        )

        for option in self.rst.get_directive_options(name):
            insert_text = f":{option}:"

            items.append(
                CompletionItem(
                    label=option,
                    detail=f"{directive.__module__}.{directive.__name__}:{option}",
                    kind=CompletionItemKind.Field,
                    filter_text=insert_text,
                    text_edit=TextEdit(range=range_, new_text=insert_text),
                    data={"completion_type": "directive_option", "for_directive": name},
                )
            )

        return items

    def completion_resolve_option(self, item: CompletionItem) -> CompletionItem:

        # We need the detail field set to the implementation's fully qualified name.
        if not item.detail or not item.data:
            return item

        directive, option = item.detail.split(":")
        name = typing.cast(Dict, item.data).get("for_directive", "")

        documentation = self.get_documentation(name, directive)
        if not documentation:
            return item

        description = documentation.get("options", {}).get(option, None)
        if not description:
            return item

        source = documentation.get("source", "")
        license = documentation.get("license", "")

        if source:
            description += f"\n\n[Source]({source})"

        if license:
            description += f"\n\n[License]({license})"

        kind = MarkupKind.PlainText
        if documentation.get("is_markdown", False):
            kind = MarkupKind.Markdown

        item.documentation = MarkupContent(kind=kind, value=description)
        return item

    definition_triggers = [DIRECTIVE]

    def definition(self, context: DefinitionContext) -> List[Location]:
        self.rst.logger.debug("%s", context)

        directive = context.match.group("name")
        domain = context.match.group("domain")
        argument = context.match.group("argument")

        if not argument:
            return []

        start = context.match.group(0).index(argument)
        end = start + len(argument)

        if start <= context.position.character <= end:
            return self.find_argument_definition(context, directive, domain, argument)

        return []

    def find_argument_definition(
        self,
        context: DefinitionContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> List[Location]:
        definitions = []

        for _, provider in self._argument_definition_providers.items():
            definitions += (
                provider.find_definitions(context, directive, domain, argument) or []
            )

        return definitions

    def get_surrounding_directive(
        self, context: CompletionContext
    ) -> Optional["re.Match"]:
        """Used to determine which directive we should be offering completions for.

        When suggestions should be generated this returns an :class:`python:re.Match`
        object representing the directive the options are associated with. In the
        case where suggestions should not be generated this will return ``None``

        Parameters
        ----------
        context:
          The completion context
        """

        match = context.match
        groups = match.groupdict()
        indent = groups["indent"]

        self.logger.debug("Match groups: %s", groups)

        # Search backwards so that we can determine the context for our completion
        linum = context.position.line - 1
        line = context.doc.lines[linum]

        while linum >= 0 and line.startswith(indent):
            linum -= 1
            line = context.doc.lines[linum]

        # Only offer completions if we're within a directive's option block
        directive = DIRECTIVE.match(line)
        self.logger.debug("Context line:  %s", line)
        self.logger.debug("Context match: %s", directive)

        if not directive:
            return None

        # Now that we know we're in a directive's option block, is the completion
        # request coming from a valid position on the line?
        option = groups["option"]
        start = match.span()[0] + match.group(0).find(option)
        end = start + len(option) + 1

        if start <= context.position.character <= end:
            return directive

        return None

    def get_documentation(
        self, label: str, implementation: str
    ) -> Optional[Dict[str, Any]]:
        """Return the documentation for the given directive, if available.

        If documentation for the given ``label`` cannot be found, this function will also
        look for the label under the project's :confval:`sphinx:primary_domain` followed
        by the ``std`` domain.

        Parameters
        ----------
        label:
           The name of the directive, as the user would type in an reStructuredText file.
        implementation:
           The full dotted name of the directive's implementation.
        """

        key = f"{label}({implementation})"
        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        if not isinstance(self.rst, SphinxLanguageServer) or not self.rst.app:
            return None

        # Nothing found, try the primary domain
        domain = self.rst.app.config.primary_domain
        key = f"{domain}:{label}({implementation})"

        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        # Still nothing, try the standard domain
        key = f"std:{label}({implementation})"

        documentation = self._documentation.get(key, None)
        if documentation:
            return documentation

        return None


def esbonio_setup(rst: RstLanguageServer):
    """Configure and reigster the directives feature with the server."""

    directives = Directives(rst)
    rst.add_feature(directives)

    docutils_docs = pkg_resources.resource_string("esbonio.lsp.rst", "directives.json")
    directives.add_documentation(json.loads(docutils_docs.decode("utf8")))

    if isinstance(rst, SphinxLanguageServer):
        sphinx_docs = pkg_resources.resource_string(
            "esbonio.lsp.sphinx", "directives.json"
        )
        directives.add_documentation(json.loads(sphinx_docs.decode("utf8")))



import argparse
import sys


def setup_cli(progname: str, description: str) -> argparse.ArgumentParser:
    """Return an argument parser with the default command line options required for
    main.

    In addition to any command line argumments you add, you should attach a function
    that takes the parsed arguments and returns a configured language server::

        cli = setup_cli("lsp-server", "my server's description")
        cli.set_defaults(configure=my_configure_function)
    """

    cli = argparse.ArgumentParser(prog=progname, description=description)
    cli.add_argument(
        "-p",
        "--port",
        type=int,
        default=None,
        help="start a TCP instance of the language server listening on the given port.",
    )
    cli.add_argument(
        "--version", action="store_true", help="print the current version and exit."
    )

    modules = cli.add_argument_group(
        "modules", "include/exclude language server modules."
    )
    modules.add_argument(
        "-i",
        "--include",
        metavar="MOD",
        action="append",
        default=[],
        dest="included_modules",
        help="include an additional module in the server configuration, can be given multiple times.",
    )
    modules.add_argument(
        "-e",
        "--exclude",
        metavar="MOD",
        action="append",
        default=[],
        dest="excluded_modules",
        help="exclude a module from the server configuration, can be given multiple times.",
    )

    return cli


def main(cli: argparse.ArgumentParser):
    """Standard main function for each of the default language servers."""

    # Put these here to avoid circular import issues.
    from esbonio.lsp import __version__
    from esbonio.lsp import create_language_server

    args = cli.parse_args()

    if args.version:
        print(f"v{__version__}")
        sys.exit(0)

    modules = list(args.modules)

    for mod in args.included_modules:
        modules.append(mod)

    for mod in args.excluded_modules:
        modules.remove(mod)

    server = create_language_server(args.server_cls, modules)

    if args.port:
        server.start_tcp("localhost", args.port)
    else:
        server.start_io()



import enum
import importlib
import json
import logging
import textwrap
import traceback
from typing import Iterable
from typing import Type

from pygls.lsp.methods import CODE_ACTION
from pygls.lsp.methods import COMPLETION
from pygls.lsp.methods import COMPLETION_ITEM_RESOLVE
from pygls.lsp.methods import DEFINITION
from pygls.lsp.methods import DOCUMENT_SYMBOL
from pygls.lsp.methods import INITIALIZE
from pygls.lsp.methods import INITIALIZED
from pygls.lsp.methods import TEXT_DOCUMENT_DID_CHANGE
from pygls.lsp.methods import TEXT_DOCUMENT_DID_OPEN
from pygls.lsp.methods import TEXT_DOCUMENT_DID_SAVE
from pygls.lsp.methods import WORKSPACE_DID_DELETE_FILES
from pygls.lsp.types import CodeActionParams
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionList
from pygls.lsp.types import CompletionOptions
from pygls.lsp.types import CompletionParams
from pygls.lsp.types import DefinitionParams
from pygls.lsp.types import DeleteFilesParams
from pygls.lsp.types import DidChangeTextDocumentParams
from pygls.lsp.types import DidOpenTextDocumentParams
from pygls.lsp.types import DidSaveTextDocumentParams
from pygls.lsp.types import DocumentSymbolParams
from pygls.lsp.types import FileOperationFilter
from pygls.lsp.types import FileOperationPattern
from pygls.lsp.types import FileOperationRegistrationOptions
from pygls.lsp.types import InitializedParams
from pygls.lsp.types import InitializeParams
from pygls.lsp.types import ServerCapabilities
from pygls.protocol import LanguageServerProtocol

from .rst import CompletionContext
from .rst import DefinitionContext
from .rst import LanguageFeature
from .rst import RstLanguageServer
from .rst import SymbolVisitor

__version__ = "0.9.0"

__all__ = [
    "LanguageFeature",
    "RstLanguageServer",
    "create_language_server",
]

logger = logging.getLogger(__name__)


class Patched(LanguageServerProtocol):
    """Tweaked version of the protocol allowing us to tweak how the `ServerCapabilities`
    are constructed."""

    def __init__(self, *args, **kwargs):
        self._server_capabilities = ServerCapabilities()
        super().__init__(*args, **kwargs)

    @property
    def server_capabilities(self):
        return self._server_capabilities

    @server_capabilities.setter
    def server_capabilities(self, value: ServerCapabilities):

        if WORKSPACE_DID_DELETE_FILES in self.fm.features:
            opts = self.fm.feature_options.get(WORKSPACE_DID_DELETE_FILES, None)
            if opts:
                value.workspace.file_operations.did_delete = opts  # type: ignore

        self._server_capabilities = value


def create_language_server(
    server_cls: Type[RstLanguageServer], modules: Iterable[str], *args, **kwargs
) -> RstLanguageServer:
    """Create a new language server instance.

    Parameters
    ----------
    server_cls:
       The class definition to create the server from.
    modules:
       The list of modules that should be loaded.
    args, kwargs:
       Any additional arguments that should be passed to the language server's
       constructor.
    """

    if "logger" not in kwargs:
        kwargs["logger"] = logger

    server = server_cls(*args, **kwargs, protocol_cls=Patched)

    for module in modules:
        _load_module(server, module)

    return _configure_lsp_methods(server)


def _configure_lsp_methods(server: RstLanguageServer) -> RstLanguageServer:
    @server.feature(INITIALIZE)
    def on_initialize(ls: RstLanguageServer, params: InitializeParams):
        ls.initialize(params)

        for feature in ls._features.values():
            feature.initialize(params)

    @server.feature(INITIALIZED)
    def on_initialized(ls: RstLanguageServer, params: InitializedParams):
        ls.initialized(params)

        for feature in ls._features.values():
            feature.initialized(params)

    @server.feature(TEXT_DOCUMENT_DID_OPEN)
    def on_open(ls: RstLanguageServer, params: DidOpenTextDocumentParams):
        pass

    @server.feature(TEXT_DOCUMENT_DID_CHANGE)
    def on_change(ls: RstLanguageServer, params: DidChangeTextDocumentParams):
        pass

    @server.feature(TEXT_DOCUMENT_DID_SAVE)
    def on_save(ls: RstLanguageServer, params: DidSaveTextDocumentParams):
        ls.save(params)

        for feature in ls._features.values():
            feature.save(params)

    @server.feature(
        WORKSPACE_DID_DELETE_FILES,
        FileOperationRegistrationOptions(
            filters=[
                FileOperationFilter(
                    pattern=FileOperationPattern(glob="**/*.rst"),
                )
            ]
        ),
    )
    def on_delete_files(ls: RstLanguageServer, params: DeleteFilesParams):
        ls.delete_files(params)

        for feature in ls._features.values():
            feature.delete_files(params)

    @server.feature(CODE_ACTION)
    def on_code_action(ls: RstLanguageServer, params: CodeActionParams):
        actions = []

        for feature in ls._features.values():
            actions += feature.code_action(params)

        return actions

    # <engine-example>
    @server.feature(
        COMPLETION,
        CompletionOptions(
            trigger_characters=[".", ":", "`", "<", "/"], resolve_provider=True
        ),
    )
    def on_completion(ls: RstLanguageServer, params: CompletionParams):
        uri = params.text_document.uri
        pos = params.position

        doc = ls.workspace.get_document(uri)
        line = ls.line_at_position(doc, pos)
        location = ls.get_location_type(doc, pos)

        items = []

        for name, feature in ls._features.items():
            for pattern in feature.completion_triggers:
                for match in pattern.finditer(line):
                    if not match:
                        continue

                    # Only trigger completions if the position of the request is within
                    # the match.
                    start, stop = match.span()
                    if start <= pos.character <= stop:
                        context = CompletionContext(
                            doc=doc, location=location, match=match, position=pos
                        )
                        ls.logger.debug("Completion context: %s", context)

                        for item in feature.complete(context):
                            item.data = {"source_feature": name, **(item.data or {})}
                            items.append(item)

        return CompletionList(is_incomplete=False, items=items)

    # </engine-example>

    @server.feature(COMPLETION_ITEM_RESOLVE)
    def on_completion_resolve(
        ls: RstLanguageServer, item: CompletionItem
    ) -> CompletionItem:
        source = (item.data or {}).get("source_feature", "")
        feature = ls.get_feature(source)

        if not feature:
            ls.logger.error(
                "Unable to resolve completion item, unknown source: '%s'", source
            )
            return item

        return feature.completion_resolve(item)

    @server.feature(DEFINITION)
    def on_definition(ls: RstLanguageServer, params: DefinitionParams):
        uri = params.text_document.uri
        pos = params.position

        doc = ls.workspace.get_document(uri)
        line = ls.line_at_position(doc, pos)
        location = ls.get_location_type(doc, pos)

        definitions = []

        for feature in ls._features.values():
            for pattern in feature.definition_triggers:
                for match in pattern.finditer(line):
                    if not match:
                        continue

                    start, stop = match.span()
                    if start <= pos.character and pos.character <= stop:
                        context = DefinitionContext(
                            doc=doc, location=location, match=match, position=pos
                        )
                        definitions += feature.definition(context)

        return definitions

    @server.feature(DOCUMENT_SYMBOL)
    def on_document_symbol(ls: RstLanguageServer, params: DocumentSymbolParams):

        doctree = ls.get_doctree(uri=params.text_document.uri)
        if doctree is None:
            return []

        visitor = SymbolVisitor(ls, doctree)
        doctree.walkabout(visitor)

        return visitor.symbols

    @server.command("esbonio.server.configuration")
    def get_configuration(ls: RstLanguageServer, *args):
        """Get the server's configuration.

        Not to be confused with the ``workspace/configuration`` request where the server
        can request the client's configuration. This is so client's can ask for sphinx's
        output path for example.

        As far as I know, there isn't anything built into the spec to cater for this?
        """
        config = ls.configuration
        ls.logger.debug("%s: %s", "esbonio.server.configuration", config)

        return config

    return server


def _load_module(server: RstLanguageServer, module: str):

    try:
        mod = importlib.import_module(module)
    except ImportError:
        logger.error("Unable to import module '%s'\n%s", module, traceback.format_exc())
        return

    if not hasattr(mod, "esbonio_setup"):
        logger.error(
            "Unable to load module '%s', missing 'esbonio_setup' function", module
        )
        return

    try:
        mod.esbonio_setup(server)
    except Exception:
        logger.error(
            "Error while setting up module '%s'\n%s", module, traceback.format_exc()
        )


def dump(obj) -> str:
    """Debug helper function that converts an object to JSON."""

    def default(o):
        if isinstance(o, enum.Enum):
            return o.value

        fields = {}
        for k, v in o.__dict__.items():

            if v is None:
                continue

            # Truncate long strings - but not uris!
            if isinstance(v, str) and not k.lower().endswith("uri"):
                v = textwrap.shorten(v, width=25)

            fields[k] = v

        return fields

    return json.dumps(obj, default=default)



"""Spell checking."""
import re
from typing import Dict
from typing import List
from typing import Optional

import pygls.uris as Uri
from docutils import nodes
from pygls.lsp.types import CodeAction
from pygls.lsp.types import CodeActionKind
from pygls.lsp.types import CodeActionParams
from pygls.lsp.types import TextEdit
from pygls.lsp.types import WorkspaceEdit
from pygls.lsp.types.basic_structures import Diagnostic
from pygls.lsp.types.basic_structures import DiagnosticSeverity
from pygls.lsp.types.basic_structures import Position
from pygls.lsp.types.basic_structures import Range
from pygls.lsp.types.workspace import DidSaveTextDocumentParams
from spellchecker import SpellChecker  # type: ignore

from esbonio.lsp.rst import LanguageFeature
from esbonio.lsp.sphinx import SphinxLanguageServer


IGNORED_NODES = {nodes.raw, nodes.literal, nodes.literal_block}
"""Don't spell check Text contained in any of these nodes."""


class Spelling(LanguageFeature):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.lang = SpellChecker()
        self.errors: Dict[str, List[MisSpelling]] = {}

    def code_action(self, params: CodeActionParams) -> List[CodeAction]:
        uri = params.text_document.uri
        ranges = {d.range for d in params.context.diagnostics}

        if uri in self.errors:
            errors = self.errors[uri]
        else:
            errors = self.find_errors_for_uri(uri)

        diagnostics = errors_to_diagnostics(errors)
        actions = []

        for error, diagnostic in zip(errors, diagnostics):
            if len(ranges) > 0 and diagnostic.range not in ranges:
                continue

            for fix in self.lang.candidates(error.text):
                actions.append(
                    CodeAction(
                        title=f"Correct '{error.text}' -> '{fix}'",
                        kind=CodeActionKind.QuickFix,
                        diagnostics=[diagnostic],
                        edit=WorkspaceEdit(
                            changes={
                                uri: [TextEdit(range=diagnostic.range, new_text=fix)]
                            }
                        ),
                    )
                )

        return actions

    def save(self, params: DidSaveTextDocumentParams):
        self.find_errors_for_uri(params.text_document.uri)

    def find_errors_for_uri(self, uri: str) -> List["MisSpelling"]:
        """Find any mis-spellings in the given document."""

        doctree = self.rst.get_doctree(uri=uri)
        if doctree is None:
            return []

        errors = []

        for text in doctree.traverse(condition=nodes.Text):
            parent = text.parent
            # Don't spell check code block, raw blocks etc.

            if type(parent) in IGNORED_NODES:
                continue

            # Don't spell check text we cannot tie back to the user's actual source file
            self.logger.debug("%s: %s", type(parent), parent.source)
            if parent.source != Uri.to_fs_path(uri):
                continue

            for word in find_words(
                text.astext(), startline=parent.line, source=parent.source
            ):
                # Ignore short words or any "word" that contains digits or other
                # punctuation.
                if len(word) <= 1 or re.search("[-/\\_\\d.=']", str(word)):
                    continue

                if self.lang.unknown([str(word)]):
                    errors.append(word)

        self.errors[uri] = errors
        self.rst.set_diagnostics("spellcheck[en]", uri, errors_to_diagnostics(errors))
        self.rst.sync_diagnostics()

        return errors


class MisSpelling:
    """Represents an incorrectly spelled word."""

    def __init__(self, line: int, character: int, text: str, source: Optional[str]):
        self.line = line
        self.character = character
        self.text = text
        self.source = source

    def __str__(self):
        return self.text

    def __contains__(self, item):
        return item in self.text

    def __len__(self):
        return len(self.text)

    def __repr__(self):
        return f"MisSpelling<{self.line}:{self.character}, {self.text}>"


def errors_to_diagnostics(errors: List[MisSpelling]) -> List[Diagnostic]:
    diagnostics = []

    for error in errors:
        range_ = Range(
            start=Position(line=error.line - 1, character=error.character),
            end=Position(line=error.line - 1, character=error.character + len(error)),
        )

        diagnostics.append(
            Diagnostic(
                range=range_,
                message=f"Incorrect spelling: '{error}'",
                severity=DiagnosticSeverity.Warning,
                source="spellcheck[en]",
            )
        )

    return diagnostics


def find_words(
    text: str, startline: int = 0, source: Optional[str] = None
) -> List[MisSpelling]:
    words = []
    delimiters = " \n"
    skip_characters = ",'\"()[]"
    current_word = None

    line = startline
    col = -1

    for c in text:
        col += 1

        if c in skip_characters:
            continue

        if c not in delimiters:
            if current_word is None:
                current_word = MisSpelling(line, col, c, source)
            else:
                current_word.text += c
        else:
            if current_word is not None:
                words.append(current_word)
                current_word = None

            if c == "\n":
                line += 1
                col = -1

    if current_word is not None:
        words.append(current_word)

    return words


def esbonio_setup(rst):

    if isinstance(rst, SphinxLanguageServer):
        spelling = Spelling(rst)
        rst.add_feature(spelling)



import pathlib
from typing import Generator

from pygls.lsp.types import CompletionItem
from pygls.lsp.types import CompletionItemKind
from pygls.lsp.types import Position
from pygls.lsp.types import Range
from pygls.lsp.types import TextEdit

from esbonio.lsp.rst import CompletionContext


def path_to_completion_item(
    context: CompletionContext, path: pathlib.Path
) -> CompletionItem:
    """Create the ``CompletionItem`` for the given path.

    In the case where there are multiple filepath components, this function needs to
    provide an appropriate ``TextEdit`` so that the most recent entry in the path can
    be easily edited - without clobbering the existing path.

    Also bear in mind that this function must play nice with both role target and
    directive argument completions.
    """

    new_text = f"{path.name}"
    kind = CompletionItemKind.Folder if path.is_dir() else CompletionItemKind.File

    # If we can't find the '/' we may as well not bother with a `TextEdit` and let the
    # `Roles` feature provide the default handling.
    start = _find_start_char(context)
    if start == -1:
        insert_text = new_text
        filter_text = None
        text_edit = None
    else:

        start += 1
        _, end = context.match.span()
        prefix = context.match.group(0)[start:]

        insert_text = None
        filter_text = (
            f"{prefix}{new_text}"  # Needed so VSCode will actually show the results.
        )

        text_edit = TextEdit(
            range=Range(
                start=Position(line=context.position.line, character=start),
                end=Position(line=context.position.line, character=end),
            ),
            new_text=new_text,
        )

    return CompletionItem(
        label=new_text,
        kind=kind,
        insert_text=insert_text,
        filter_text=filter_text,
        text_edit=text_edit,
    )


def _find_start_char(context: CompletionContext) -> int:
    matched_text = context.match.group(0)
    idx = matched_text.find("/")

    while True:
        next_idx = matched_text.find("/", idx + 1)
        if next_idx == -1:
            break

        idx = next_idx

    return idx


def complete_filepaths(base: str, partial: str) -> Generator[pathlib.Path, None, None]:
    """Generate filepath completion suggestions relative to the given base.

    This function is for "docutils style" behaviour where the path is relative to the
    current document.

    Parameters
    ----------
    base
       The directory containing the current document

    partial
       The existing path entered so far.
    """

    candidate_dir = pathlib.Path(base) / pathlib.Path(partial)
    if partial and not partial.endswith("/"):
        candidate_dir = candidate_dir.parent

    return candidate_dir.glob("*")


def complete_sphinx_filepaths(
    srcdir: str, base: str, partial: str
) -> Generator[pathlib.Path, None, None]:
    """Generate filepath completion suggestions relative to the given base or ``srcdir``.

    This function is for "sphinx style" behaviour where the path is relative to the
    current document *unless* ``partial`` starts with a ``/`` character. In this case
    completions should be relative to the given ``srcdir``.

    Parameters
    ----------
    srcdir
       The ``srcdir`` of the project, used when ``partial`` starts with a ``/``.

    base
       The directory containing the current document.

    partial
       The existing path entered so far.
    """

    if partial and partial.startswith("/"):
        candidate_dir = pathlib.Path(srcdir)

        # Be sure to take off the leading '/' character, otherwise the partial
        # path will wipe out the srcdir when concatenated...
        partial = partial[1:]
    else:
        candidate_dir = pathlib.Path(base)

    candidate_dir /= pathlib.Path(partial)
    if partial and not partial.endswith("/"):
        candidate_dir = candidate_dir.parent

    return candidate_dir.glob("*")



"""Default entry point, identical to calling ``python -m esbonio.lsp.sphinx``"""
import sys

from esbonio.cli import main
from esbonio.lsp.sphinx import cli

if __name__ == "__main__":
    sys.exit(main(cli))



"""Extra support for roles added by sphinx."""
import os.path
import typing
from typing import List
from typing import Optional

import pygls.uris as Uri
from pygls.lsp.types import CompletionItem

from esbonio.lsp.roles import Roles
from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.sphinx import SphinxLanguageServer
from esbonio.lsp.util.filepaths import complete_sphinx_filepaths
from esbonio.lsp.util.filepaths import path_to_completion_item


class Downloads:
    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)

    def complete_targets(
        self, context: CompletionContext, name: str, domain: Optional[str]
    ) -> List[CompletionItem]:

        if domain or name != "download":
            return []

        if not self.rst.app:
            return []

        srcdir = self.rst.app.srcdir
        partial = context.match.group("label")
        base = os.path.dirname(Uri.to_fs_path(context.doc.uri))
        items = complete_sphinx_filepaths(srcdir, base, partial)

        return [path_to_completion_item(context, p) for p in items]


def esbonio_setup(rst: RstLanguageServer):

    name = "esbonio.lsp.roles.Roles"
    roles = rst.get_feature(name)

    if not isinstance(rst, SphinxLanguageServer) or not roles:
        return

    # To keep mypy happy
    roles = typing.cast(Roles, roles)
    roles.add_target_completion_provider(Downloads(rst))



import os.path
import pathlib
import typing
from typing import List
from typing import Optional

import pygls.uris as Uri
from pygls.lsp.types import CompletionItem
from pygls.lsp.types import Location
from pygls.lsp.types import Position
from pygls.lsp.types import Range

from esbonio.lsp.directives import Directives
from esbonio.lsp.rst import CompletionContext
from esbonio.lsp.rst import DefinitionContext
from esbonio.lsp.rst import RstLanguageServer
from esbonio.lsp.sphinx import SphinxLanguageServer
from esbonio.lsp.util.filepaths import complete_sphinx_filepaths
from esbonio.lsp.util.filepaths import path_to_completion_item


class Includes:
    def __init__(self, rst: SphinxLanguageServer):
        self.rst = rst
        self.logger = rst.logger.getChild(self.__class__.__name__)

    def complete_arguments(
        self, context: CompletionContext, domain: str, name: str
    ) -> List[CompletionItem]:

        if domain or name not in {"include", "literalinclude"}:
            return []

        if not self.rst.app:
            return []

        srcdir = self.rst.app.srcdir
        partial = context.match.group("argument")
        base = os.path.dirname(Uri.to_fs_path(context.doc.uri))
        items = complete_sphinx_filepaths(srcdir, base, partial)

        return [path_to_completion_item(context, p) for p in items]

    def find_definitions(
        self,
        context: DefinitionContext,
        directive: str,
        domain: Optional[str],
        argument: str,
    ) -> List[Location]:

        if domain or directive not in {"literalinclude", "include"}:
            return []

        if argument.startswith("/"):
            if not self.rst.app:
                return []

            basedir = pathlib.Path(self.rst.app.srcdir)

            # Remove the leading '/' otherwise is will wipe out the basedir when
            # concatenated
            argument = argument[1:]

        else:
            basedir = pathlib.Path(Uri.to_fs_path(context.doc.uri)).parent

        fpath = (basedir / argument).resolve()
        if not fpath.exists():
            return []

        return [
            Location(
                uri=Uri.from_fs_path(str(fpath)),
                range=Range(
                    start=Position(line=0, character=0),
                    end=Position(line=1, character=0),
                ),
            )
        ]


def esbonio_setup(rst: RstLanguageServer):

    name = "esbonio.lsp.directives.Directives"
    directives = rst.get_feature(name)
    if not isinstance(rst, SphinxLanguageServer) or not directives:
        return

    # To keep mypy happy.
    directives = typing.cast(Directives, directives)

    includes = Includes(rst)
    directives.add_argument_definition_provider(includes)
    directives.add_argument_completion_provider(includes)
